@@ -23,6 +23,7 @@ from .api import (
   ad,  # TODO(phawkins): update users to avoid this.
   argnums_partial,  # TODO(phawkins): update Haiku to not use this.
   checkpoint,
+  closure_convert,
   curry,  # TODO(phawkins): update users to avoid this.
   custom_ivjp,
   custom_gradient,
@@ -36,6 +37,8 @@ from .api import (
   device_count,
   device_get,
   device_put,
+  device_put_sharded,
+  device_put_replicated,
   devices,
   disable_jit,
   eval_shape,
@@ -92,6 +95,7 @@ from . import lax
 from . import nn
 from . import profiler
 from . import random
+from . import util
 
 def _init():
   from . import numpy # side-effecting import sets up operator overloads
