@@ -16,6 +16,7 @@ Parallelization primitives.
 """
 
 import collections
+import string
 import warnings
 
 import numpy as np
@@ -25,13 +26,13 @@ from jax import dtypes
 from jax import tree_util
 from jax._src import source_info_util
 from . import lax
-from jax.abstract_arrays import ShapedArray, raise_to_shaped
+from jax.core import ShapedArray, raise_to_shaped
 from jax.interpreters import ad
 from jax.interpreters import xla
 from jax.interpreters import pxla
 from jax.interpreters import batching
 from jax.interpreters import partial_eval as pe
-from jax.util import partial, unzip2, prod
+from jax._src.util import partial, unzip2, prod
 from jax.lib import xla_client as xc
 from jax.lib import xla_bridge as xb
 from jax.config import config
@@ -74,6 +75,8 @@ def psum(x, axis_name, *, axis_index_groups=None):
   >>> print(y)
   [ 0.          0.16666667  0.33333334  0.5       ]
   """
+  if not isinstance(axis_name, (tuple, list)):
+    axis_name = (axis_name,)
   _validate_axis_index_groups(axis_index_groups)
   leaves, treedef = tree_util.tree_flatten(x)
   leaves = [lax.convert_element_type(l, np.int32)
@@ -134,9 +137,13 @@ def pmax(x, axis_name, *, axis_index_groups=None):
     Array(s) with the same shape as ``x`` representing the result of an
     all-reduce max along the axis ``axis_name``.
   """
+  if not isinstance(axis_name, (tuple, list)):
+    axis_name = (axis_name,)
   _validate_axis_index_groups(axis_index_groups)
-  return tree_util.tree_map(partial(
-      pmax_p.bind, axis_name=axis_name, axis_index_groups=axis_index_groups), x)
+  leaves, treedef = tree_util.tree_flatten(x)
+  out_flat = pmax_p.bind(*leaves, axis_name=axis_name,
+                         axis_index_groups=axis_index_groups)
+  return tree_util.tree_unflatten(treedef, out_flat)
 
 def pmin(x, axis_name, *, axis_index_groups=None):
   """Compute an all-reduce min on ``x`` over the pmapped axis ``axis_name``.
@@ -157,9 +164,13 @@ def pmin(x, axis_name, *, axis_index_groups=None):
     Array(s) with the same shape as ``x`` representing the result of an
     all-reduce min along the axis ``axis_name``.
   """
+  if not isinstance(axis_name, (tuple, list)):
+    axis_name = (axis_name,)
   _validate_axis_index_groups(axis_index_groups)
-  return tree_util.tree_map(partial(
-      pmin_p.bind, axis_name=axis_name, axis_index_groups=axis_index_groups), x)
+  leaves, treedef = tree_util.tree_flatten(x)
+  out_flat = pmin_p.bind(*leaves, axis_name=axis_name,
+                         axis_index_groups=axis_index_groups)
+  return tree_util.tree_unflatten(treedef, out_flat)
 
 def _validate_axis_index_groups(axis_index_groups):
   if axis_index_groups is None:
@@ -223,14 +234,16 @@ def pshuffle(x, axis_name, perm):
   return ppermute(x, axis_name, list(zip(perm, range(len(perm)))))
 
 
-def pswapaxes(x, axis_name, axis):
+def pswapaxes(x, axis_name, axis, *, axis_index_groups=None):
   """Swap the pmapped axis ``axis_name`` with the unmapped axis ``axis``.
 
   If ``x`` is a pytree then the result is equivalent to mapping this function to
   each leaf in the tree.
 
-  The mapped axis size must be equal to the size of the unmapped axis; that is,
-  we must have ``lax.psum(1, axis_name) == x.shape[axis]``.
+  The group size of the mapped axis size must be equal to the size of the
+  unmapped axis; that is, we must have
+  ``lax.psum(1, axis_name, axis_index_groups=axis_index_groups) == x.shape[axis]``.
+  By default, when ``axis_index_groups=None``, this encompasses all the devices.
 
   This function is a special case of ``all_to_all`` where the pmapped axis of
   the input is placed at the position ``axis`` in the output. That is, it is
@@ -242,13 +255,17 @@ def pswapaxes(x, axis_name, axis):
       :func:`jax.pmap` documentation for more details).
     axis: int indicating the unmapped axis of ``x`` to map with the name
       ``axis_name``.
+    axis_index_groups: optional list of lists containing axis indices (e.g. for
+      an axis of size 4, [[0, 1], [2, 3]] would run pswapaxes over the first
+      two and last two replicas). Groups must cover all axis indices exactly
+      once, and all groups must be the same size.
 
   Returns:
     Array(s) with the same shape as ``x``.
   """
-  return all_to_all(x, axis_name, axis, axis)
+  return all_to_all(x, axis_name, axis, axis, axis_index_groups=axis_index_groups)
 
-def all_to_all(x, axis_name, split_axis, concat_axis):
+def all_to_all(x, axis_name, split_axis, concat_axis, *, axis_index_groups=None):
   """Materialize the mapped axis and map a different axis.
 
   If ``x`` is a pytree then the result is equivalent to mapping this function to
@@ -258,8 +275,10 @@ def all_to_all(x, axis_name, split_axis, concat_axis):
   logical axis position ``concat_axis``, and the input unmapped axis at position
   ``split_axis`` is mapped with the name ``axis_name``.
 
-  The input mapped axis size must be equal to the size of the axis to be mapped;
-  that is, we must have ``lax.psum(1, axis_name) == x.shape[split_axis]``.
+  The group size of the mapped axis size must be equal to the size of the
+  unmapped axis; that is, we must have
+  ``lax.psum(1, axis_name, axis_index_groups=axis_index_groups) == x.shape[axis]``.
+  By default, when ``axis_index_groups=None``, this encompasses all the devices.
 
   Args:
     x: array(s) with a mapped axis named ``axis_name``.
@@ -269,6 +288,10 @@ def all_to_all(x, axis_name, split_axis, concat_axis):
       ``axis_name``.
     concat_axis: int indicating the position in the output to materialize the
       mapped axis of the input with the name ``axis_name``.
+    axis_index_groups: optional list of lists containing axis indices (e.g. for
+      an axis of size 4, [[0, 1], [2, 3]] would run all_to_all over the first
+      two and last two replicas). Groups must cover all axis indices exactly
+      once, and all groups must be the same size.
 
   Returns:
     Array(s) with shape given by the expression::
@@ -279,12 +302,15 @@ def all_to_all(x, axis_name, split_axis, concat_axis):
     the input ``x``, i.e. ``axis_size = lax.psum(1, axis_name)``.
   """
   def bind(x):
-    if psum(1, axis_name) != x.shape[split_axis]:
+    group_size = psum(1, axis_name, axis_index_groups=axis_index_groups)
+    if group_size != x.shape[split_axis]:
       msg = ("all_to_all requires the size of the mapped axis axis_name to "
              "equal x.shape[split_axis], but they are {} and {} respectively.")
-      raise ValueError(msg.format(psum(1, axis_name), x.shape[split_axis]))
+      raise ValueError(msg.format(group_size, x.shape[split_axis]))
     return all_to_all_p.bind(x, split_axis=split_axis, concat_axis=concat_axis,
-                             axis_name=axis_name)
+                             axis_name=axis_name,
+                             axis_index_groups=axis_index_groups)
+
   return tree_util.tree_map(bind, x)
 
 def axis_index(axis_name):
@@ -327,6 +353,135 @@ def axis_index(axis_name):
   return axis_index_p.bind(axis_name=axis_name)
 
 
+def pdot(x, y, axis_name, pos_contract=((), ()), pos_batch=((), ())):
+  if not isinstance(axis_name, (list, tuple)):
+    axis_name = (axis_name,)
+  return pdot_p.bind(x, y, axis_name=axis_name,
+                     pos_contract=pos_contract, pos_batch=pos_batch)
+
+
+def xeinsum(spec: str, x, y):
+  in_spec, out_spec = spec.split('->')
+  (lhs_subs, lhs_named), (rhs_subs, rhs_named) = XeinsumSpecParser(in_spec).parse_args()
+  (out_subs, out_named), = XeinsumSpecParser(out_spec).parse_args()
+  all_named = {*lhs_named, *rhs_named, *out_named}
+  all_subs = {*lhs_subs, *rhs_subs, *out_subs}
+  lhs_uniques = set(lhs_subs) - set(rhs_subs)
+  rhs_uniques = set(rhs_subs) - set(lhs_subs)
+  if all_subs & all_named:
+    raise NotImplementedError
+  if not set(out_named).issubset({*lhs_named, *rhs_named}):
+    raise ValueError
+
+  # if a named axis appears in both inputs and not the output, contract!
+  named_contract = list(all_named - set(out_named))
+
+  # if a subscript appears in both inputs and not the outputs, contract!
+  subs_contract = all_subs - set(out_subs)
+
+  lhs_reduce_axes = [lhs_subs.index(n) for n in lhs_uniques & subs_contract]
+  if lhs_reduce_axes:
+    x = lax._reduce_sum(x, lhs_reduce_axes)
+    for i in sorted(lhs_reduce_axes, reverse=True):
+      del lhs_subs[i]
+
+  rhs_reduce_axes = [rhs_subs.index(n) for n in rhs_uniques & subs_contract]
+  if rhs_reduce_axes:
+    y = lax._reduce_sum(y, rhs_reduce_axes)
+    for i in sorted(rhs_reduce_axes, reverse=True):
+      del rhs_subs[i]
+
+  pos_contract = unzip2((lhs_subs.index(n), rhs_subs.index(n))
+                        for n in subs_contract - (lhs_uniques | rhs_uniques))
+
+  # if a subscript apperas in both inputs _and_ the outputs, batch!
+  subs_batch = all_subs - subs_contract
+  if subs_batch & (lhs_uniques | rhs_uniques):
+    raise NotImplementedError
+
+  pos_batch = unzip2((lhs_subs.index(n), rhs_subs.index(n))
+                        for n in subs_batch)
+
+  return pdot(x, y, axis_name=named_contract,
+              pos_contract=pos_contract, pos_batch=pos_batch)
+
+class XeinsumSpecParser:
+  spec: str
+  pos: int
+
+  def __init__(self, spec: str):
+    self.spec = spec
+    self.pos = 0
+
+  @property
+  def eof(self):
+    return self.pos == len(self.spec)
+
+  @property
+  def cur(self):
+    return self.spec[self.pos]
+
+  def parse_subscript(self):
+    if self.cur in string.ascii_lowercase:
+      out = self.cur
+      self.pos += 1
+      return out, True
+    else:
+      return None, False
+
+  def parse_axis_name(self):
+    try:
+      end = self.spec.index('}', self.pos)
+    except ValueError:
+      assert False
+
+    try:
+      end = self.spec.index(',', self.pos, end)
+    except ValueError:
+      pass
+
+    axis_name = self.spec[self.pos:end]
+    assert axis_name
+    self.pos = end + 1
+    return axis_name, self.spec[end] == ','
+
+  def maybe_take(self, char: str, on_eof: bool = False):
+    if self.eof:
+      return on_eof
+    if self.cur == char:
+      self.pos += 1
+      return True
+
+  def parse_arg(self):
+    subscripts = []
+    names = []
+    while not self.eof:
+      subscript, cont = self.parse_subscript()
+      if not cont: break
+      subscripts.append(subscript)
+    if self.eof:
+      return False, (subscripts, names)
+    if self.maybe_take(','):
+      return True, (subscripts, names)
+    else:
+      assert self.maybe_take('{')
+      while True:
+        axis_name, cont = self.parse_axis_name()
+        names.append(axis_name)
+        if not cont: break
+      return self.maybe_take(',', False), (subscripts, names)
+
+  def parse_args(self):
+    arg_specs = []
+    cont = True
+    while not self.eof:
+      cont, result = self.parse_arg()
+      arg_specs.append(result)
+    if cont:
+      arg_specs.append(([], []))
+    return arg_specs
+
+
 ### parallel primitives
 
 def _allreduce_soft_pmap_rule(prim, reducer, vals, mapped, chunk_size,
@@ -338,51 +493,26 @@ def _allreduce_soft_pmap_rule(prim, reducer, vals, mapped, chunk_size,
                    axis_index_groups=axis_index_groups)
   return outs, (False,) * len(vals)
 
-def _allreduce_translation_rule(prim, c, val, *, axis_name, axis_index_groups,
-                                axis_env, platform):
-  replica_groups = _replica_groups(axis_env, axis_name, axis_index_groups)
-  dtype = c.get_shape(val).numpy_dtype()
-  scalar = ShapedArray((), dtype)
-  computation = xla.primitive_subcomputation(prim, scalar, scalar)
-  replica_groups_protos = xc.make_replica_groups(replica_groups)
-  return xops.AllReduce(val, computation, replica_groups_protos, None, None)
-
-# It is assumed that all collectives that use this rule are commutative
-# and associative over axis names if they support tuples. That is,
-# they have to satisfy:
-#   collective(x, ('i', 'j')) == collective(x, ('j', 'i'))
-#                             == collective(collective(x, 'j'), 'i')
-def _split_axis_comm_assoc(primitive, split_name, args, params):
-  axis_names = params['axis_name']
-  assert isinstance(axis_names, tuple)
-  if params['axis_index_groups'] is not None:
-    raise NotImplementedError("axis_index_groups not supported in axis splitting. "
-                              "Please open a feature request!")
-  remaining_axes = list(axis_names)
-  remaining_axes.remove(split_name)
-  remaining_axes = tuple(remaining_axes)
-  split_params = dict(params, axis_name=split_name)
-  remain_params = dict(params, axis_name=remaining_axes)
-  split_result = primitive.bind(*args, **split_params)
-  if not primitive.multiple_results:
-    split_result = (split_result,)
-  return primitive.bind(*split_result, **remain_params)
-
-# NB: This is only used for collectives that do not include the vmapped axis name,
-#     which is why the rule is so simple. All other collectives go through split_axis.
+# This is only used for collectives that do not include the vmapped axis name,
+# which is why the rule is so simple.
 def _collective_batcher(prim, args, dims, **params):
   return prim.bind(*args, **params), dims if prim.multiple_results else dims[0]
 
-def _batched_reduction_collective(prim, if_mapped, if_unmapped,
-                                  vals_in, dims_in, axis_size,
-                                  axis_name, axis_index_groups):
+def _batched_reduction_collective(
+    prim, if_mapped, if_unmapped, frame, vals_in, dims_in, axis_name,
+    axis_index_groups):
+  assert prim.multiple_results
+  assert frame.name in axis_name
   if axis_index_groups is not None:
-    raise NotImplementedError("axis_index_groups not implemented in vmap collectives. "
+    raise NotImplementedError("axis_index_groups not supported in vmap collectives. "
                               "Please open a feature request!")
-  vals_out = [if_mapped(v, d) if d is not batching.not_mapped else if_unmapped(v, axis_size)
-              for v, d in zip(vals_in, dims_in)]
-  dims_out = [batching.not_mapped] * len(vals_in)
-  return vals_out, dims_out
+  vals_out = [if_mapped(v, d) if d is not batching.not_mapped
+              else if_unmapped(v, frame.size) for v, d in zip(vals_in, dims_in)]
+  if len(axis_name) > 1:
+    remaining_axis_names = tuple(n for n in axis_name if n != frame.name)
+    vals_out = prim.bind(*vals_out, axis_name=remaining_axis_names,
+                         axis_index_groups=None)
+  return vals_out, [batching.not_mapped] * len(vals_out)
 
 def _replica_groups(axis_env, axis_name, axis_index_groups):
   replica_groups = xla.axis_groups(axis_env, axis_name)
@@ -392,13 +522,13 @@ def _replica_groups(axis_env, axis_name, axis_index_groups):
                       for axis_index_group in axis_index_groups]
   return replica_groups
 
-# psum translation rule has special handling for complex dtypes
-def _psum_translation_rule(c, *args, axis_name, axis_index_groups, axis_env,
-                           platform):
+def _allreduce_translation_rule(prim, c, *args, axis_name, axis_index_groups,
+                                axis_env, platform):
   if platform in ("cpu", "tpu"):
-    return _notuple_psum_translation_rule(c, *args, axis_name=axis_name,
-                                          axis_index_groups=axis_index_groups,
-                                          axis_env=axis_env, platform=platform)
+    return _notuple_allreduce_translation_rule(
+        prim, c, *args, axis_name=axis_name,
+        axis_index_groups=axis_index_groups, axis_env=axis_env,
+        platform=platform)
 
   # XLA's tuple all-reduce doesn't support different dtypes in the same
   # allreduce. Instead, we perform once all-reduce for each argument input type.
@@ -415,14 +545,16 @@ def _psum_translation_rule(c, *args, axis_name, axis_index_groups, axis_env,
   for dtype, (indices, dtype_args) in sorted(args_by_type.items()):
     is_complex = dtypes.issubdtype(dtype, np.complexfloating)
     n = len(dtype_args)
-    if is_complex:
+    if is_complex and prim is lax.add_p:
+      # TODO(b/141575627): we handle complex-dtype sum-reduction directly as a
+      # special case because it's not currently handled by XLA:GPU
       dtype_args = ([xops.Real(x) for x in dtype_args] +
                     [xops.Imag(x) for x in dtype_args])
     scalar = ShapedArray((), c.get_shape(dtype_args[0]).numpy_dtype())
-    computation = xla.primitive_subcomputation(lax.add_p, scalar, scalar)
+    computation = xla.primitive_subcomputation(prim, scalar, scalar)
     all_reduce = xops.AllReduce(xops.Tuple(c, dtype_args), computation,
                                 replica_groups_protos, None, None)
-    if is_complex:
+    if is_complex and prim is lax.add_p:
       xs = [xops.Complex(xops.GetTupleElement(all_reduce, i),
                          xops.GetTupleElement(all_reduce, n + i)) for i in range(n)]
     else:
@@ -431,25 +563,28 @@ def _psum_translation_rule(c, *args, axis_name, axis_index_groups, axis_env,
       out[i] = x
   return xops.Tuple(c, out)
 
-# TODO(b/150476027): CPU doesn't support tuple all-reduce correctly. But
-# fortunately we don't really need it in that case because CPU doesn't support
-# cross-task communication either.
 # TODO(b/155446630): An XLA:TPU optimization pass also doesn't support
 # tuple all-reduce yet. Meanwhile, rely on deterministic compiler behavior.
-def _notuple_psum_translation_rule(c, *args, axis_name, axis_env,
-                                   axis_index_groups, platform):
-  def _translate(val):
-    psum = partial(_allreduce_translation_rule, lax.add_p, c,
-                   axis_name=axis_name, axis_env=axis_env,
-                   axis_index_groups=axis_index_groups, platform=platform)
-    dtype = c.get_shape(val).numpy_dtype()
-    if dtypes.issubdtype(dtype, np.complexfloating):
-      return xops.Complex(psum(xops.Real(val)), psum(xops.Imag(val)))
-    else:
-      return psum(val)
-  return xops.Tuple(c, list(map(_translate, args)))
+def _notuple_allreduce_translation_rule(prim, c, *args, axis_name, axis_env,
+                                        axis_index_groups, platform):
+  def all_reduce(x):
+    replica_groups_protos = xc.make_replica_groups(
+        _replica_groups(axis_env, axis_name, axis_index_groups))
+    scalar = ShapedArray((), c.get_shape(x).numpy_dtype())
+    computation = xla.primitive_subcomputation(prim, scalar, scalar)
+    return xops.AllReduce(x, computation, replica_groups_protos, None, None)
 
-def _psum_transpose_rule(cts, axis_name, axis_index_groups):
+  if prim is not lax.add_p:
+    outs = [all_reduce(x) for x in args]
+  else:
+    # TODO(b/141575627): we handle complex-dtype sum-reduction directly as a
+    # special case because it's not currently handled by XLA:GPU
+    outs = [xops.Complex(all_reduce(xops.Real(x)), all_reduce(xops.Imag(x)))
+            if dtypes.issubdtype(c.get_shape(x).numpy_dtype(), np.complexfloating)
+            else all_reduce(x) for x in args]
+  return xops.Tuple(c, outs)
+
+def _psum_transpose_rule(cts, *args, axis_name, axis_index_groups):
   nonzero_out_cts, treedef = tree_util.tree_flatten(cts)
   nonzero_in_cts = psum_p.bind(*nonzero_out_cts, axis_name=axis_name,
                                axis_index_groups=axis_index_groups)
@@ -460,10 +595,9 @@ psum_p.multiple_results = True
 psum_p.def_abstract_eval(lambda *args, **params: map(raise_to_shaped, args))
 pxla.soft_pmap_rules[psum_p] = \
     partial(_allreduce_soft_pmap_rule, psum_p, lax._reduce_sum)
-xla.parallel_translations[psum_p] = _psum_translation_rule
-ad.deflinear(psum_p, _psum_transpose_rule)
+xla.parallel_translations[psum_p] = partial(_allreduce_translation_rule, lax.add_p)  # type: ignore
+ad.deflinear2(psum_p, _psum_transpose_rule)
 pxla.multi_host_supported_collectives.add(psum_p)
-batching.split_axis_rules[psum_p] = partial(_split_axis_comm_assoc, psum_p)
 batching.primitive_batchers[psum_p] = partial(_collective_batcher, psum_p)
 batching.collective_rules[psum_p] = \
   partial(_batched_reduction_collective,
@@ -478,7 +612,7 @@ def psum_bind(*args, axis_name, axis_index_groups):
   if all(not isinstance(x, core.Tracer) for x in args):
     if axis_index_groups is not None:
       size = len(axis_index_groups[0])
-    elif type(axis_name) is tuple:
+    elif isinstance(axis_name, (list, tuple)):
       size = prod([core.axis_frame(name).size for name in axis_name])  # type: ignore
     else:
       size = core.axis_frame(axis_name).size  # type: ignore
@@ -488,31 +622,25 @@ def psum_bind(*args, axis_name, axis_index_groups):
 
 
 pmax_p = core.Primitive('pmax')
-pmax_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
-xla.parallel_translations[pmax_p] = \
-    partial(_allreduce_translation_rule, lax.max_p)
+pmax_p.multiple_results = True
+pmax_p.def_abstract_eval(lambda *args, **params: map(raise_to_shaped, args))
+xla.parallel_translations[pmax_p] = partial(_allreduce_translation_rule, lax.max_p)
 pxla.multi_host_supported_collectives.add(pmax_p)
-batching.split_axis_rules[pmax_p] = partial(_split_axis_comm_assoc, pmax_p)
 batching.primitive_batchers[pmax_p] = partial(_collective_batcher, pmax_p)
 batching.collective_rules[pmax_p] = \
-  partial(_batched_reduction_collective,
-          pmax_p,
-          lambda v, d: v.max(d),
-          lambda v, axis_size: v)
+  partial(_batched_reduction_collective, pmax_p,
+          lambda v, d: v.max(d), lambda v, axis_size: v)
 
 
 pmin_p = core.Primitive('pmin')
-pmin_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
-xla.parallel_translations[pmin_p] = \
-    partial(_allreduce_translation_rule, lax.min_p)
+pmin_p.multiple_results = True
+pmin_p.def_abstract_eval(lambda *args, **params: map(raise_to_shaped, args))
+xla.parallel_translations[pmin_p] = partial(_allreduce_translation_rule, lax.min_p)
 pxla.multi_host_supported_collectives.add(pmin_p)
-batching.split_axis_rules[pmin_p] = partial(_split_axis_comm_assoc, pmin_p)
 batching.primitive_batchers[pmin_p] = partial(_collective_batcher, pmin_p)
 batching.collective_rules[pmin_p] = \
-  partial(_batched_reduction_collective,
-          pmin_p,
-          lambda v, d: v.min(d),
-          lambda v, axis_size: v)
+  partial(_batched_reduction_collective, pmin_p,
+          lambda v, d: v.min(d), lambda v, axis_size: v)
 
 
 def _ppermute_translation_rule(c, x, *, axis_name, axis_env, perm, platform):
@@ -529,23 +657,24 @@ def _ppermute_translation_rule(c, x, *, axis_name, axis_env, perm, platform):
     full_perm.extend((grp[src], grp[dst]) for src, dst in perm)
   return xops.CollectivePermute(x, full_perm)
 
-def _ppermute_transpose_rule(t, perm, axis_name):
+def _ppermute_transpose_rule(t, x, perm, axis_name):
   srcs, dsts = unzip2(perm)
   inverse_perm = list(zip(dsts, srcs))
   return [ppermute(t, axis_name=axis_name, perm=inverse_perm)]
 
-def _ppermute_batcher(vals_in, dims_in, axis_size, axis_name, perm):
-  assert len(perm) == axis_size, "Permutation doesn't match the axis size!"
-  perm_indices = np.full((axis_size,), -1, dtype=np.int32)
-  for s, d in perm:
-    perm_indices[s] = d
-  vals_out = [lax_numpy.take(v, perm_indices, d) if d is not batching.not_mapped else v
-              for v, d in zip(vals_in, dims_in)]
-  return vals_out, dims_in
+def _ppermute_batcher(frame, vals_in, dims_in, axis_name, perm):
+  assert len(perm) == frame.size, "Permutation doesn't match the axis size!"
+  assert axis_name == frame.name, "ppermute batcher called with wrong axis name"
+  (v,), (d,) = vals_in, dims_in
+  assert d is not batching.not_mapped
+  perm_indices = [None] * frame.size
+  for src, dst in perm:
+    perm_indices[src] = dst
+  return lax_numpy.take(v, perm_indices, d), d
 
 ppermute_p = core.Primitive('ppermute')
 ppermute_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
-ad.deflinear(ppermute_p, _ppermute_transpose_rule)
+ad.deflinear2(ppermute_p, _ppermute_transpose_rule)
 xla.parallel_translations[ppermute_p] = _ppermute_translation_rule
 pxla.multi_host_supported_collectives.add(ppermute_p)
 batching.primitive_batchers[ppermute_p] = partial(_collective_batcher, ppermute_p)
@@ -557,16 +686,18 @@ def _moveaxis(src, dst, x):
   perm.insert(dst, src)
   return lax.transpose(x, perm)
 
-def _all_to_all_via_all_gather(x, *, axis_name, split_axis, concat_axis):
-  global_full = all_gather(x, axis_name)
+def _all_to_all_via_all_gather(x, *, axis_name, split_axis, concat_axis, axis_index_groups):
+  global_full = all_gather(x, axis_name, axis_index_groups=axis_index_groups)
   idx = axis_index(axis_name)
+  if axis_index_groups:
+    idx = idx % len(axis_index_groups[0])
   local_slice = lax.dynamic_index_in_dim(global_full, idx, split_axis + 1, keepdims=False)
   return _moveaxis(0, concat_axis, local_slice)
 
 def _all_to_all_translation_rule(c, x, *, split_axis, concat_axis, axis_name,
-                                 axis_env, platform):
+                                 axis_index_groups, axis_env, platform):
   # Workaround for AllToAll not being implemented on CPU.
-  replica_groups = _replica_groups(axis_env, axis_name, None)
+  replica_groups = _replica_groups(axis_env, axis_name, axis_index_groups)
   if len(replica_groups[0]) == 1:
     return x
   elif platform != 'tpu':
@@ -576,7 +707,7 @@ def _all_to_all_translation_rule(c, x, *, split_axis, concat_axis, axis_name,
     lowering = xla.lower_fun(_all_to_all_via_all_gather, multiple_results=False, parallel=True)
     return lowering(c, x,
                     split_axis=split_axis, concat_axis=concat_axis, axis_name=axis_name,
-                    axis_env=axis_env, platform=platform)
+                    axis_index_groups=axis_index_groups, axis_env=axis_env, platform=platform)
   else:
     split_count = len(replica_groups[0])
     if not all(split_count == len(g) for g in replica_groups):
@@ -595,10 +726,16 @@ def _all_to_all_translation_rule(c, x, *, split_axis, concat_axis, axis_name,
       x = xla.lower_fun(partial(lax.squeeze, dimensions=(split_axis,)), multiple_results=False)(c, x)
       return x
 
-def _all_to_all_transpose_rule(cts, axis_name, split_axis, concat_axis):
-  return (all_to_all(cts, axis_name=axis_name, split_axis=concat_axis, concat_axis=split_axis),)
+def _all_to_all_transpose_rule(cts, x, axis_name, split_axis, concat_axis, axis_index_groups):
+  return (all_to_all(
+      cts,
+      axis_name=axis_name,
+      split_axis=concat_axis,
+      concat_axis=split_axis,
+      axis_index_groups=axis_index_groups),)
 
-def _all_to_all_batcher(vals_in, dims_in, *, axis_name, split_axis, concat_axis):
+
+def _all_to_all_batcher(vals_in, dims_in, *, axis_name, split_axis, concat_axis, axis_index_groups):
   x, = vals_in
   d, = dims_in
   if d <= split_axis:
@@ -611,10 +748,19 @@ def _all_to_all_batcher(vals_in, dims_in, *, axis_name, split_axis, concat_axis)
     d -= 1
   elif concat_axis < d < split_axis:
     d += 1
-  result = all_to_all_p.bind(x, axis_name=axis_name, split_axis=split_axis, concat_axis=concat_axis)
+  result = all_to_all_p.bind(
+      x,
+      axis_name=axis_name,
+      split_axis=split_axis,
+      concat_axis=concat_axis,
+      axis_index_groups=axis_index_groups)
   return result, d
 
-def _all_to_all_batched_collective(vals_in, dims_in, axis_size, axis_name, split_axis, concat_axis):
+def _all_to_all_batched_collective(frame, vals_in, dims_in,
+                                   axis_name, split_axis, concat_axis,
+                                   axis_index_groups):
+  if isinstance(axis_name, (list, tuple)) and len(axis_name) > 1:
+    raise NotImplementedError("update after #4835")  # TODO(mattjj,apaszke)
   x, = vals_in
   d, = dims_in
   split_axis_adj = split_axis + (1 if d <= split_axis else 0)
@@ -623,42 +769,9 @@ def _all_to_all_batched_collective(vals_in, dims_in, axis_size, axis_name, split
     split_axis_adj -= 1
   elif concat_axis_adj < split_axis_adj < d:
     split_axis_adj += 1
-  return [_moveaxis(d, concat_axis_adj, x)], [split_axis_adj]
+  return _moveaxis(d, concat_axis_adj, x), split_axis_adj
 
-def _all_to_all_split_axis_rule(split_name, vals, params):
-  concat_axis = params['concat_axis']
-  split_axis = params['split_axis']
-  axis_names = params['axis_name']
-  assert isinstance(axis_names, tuple)
-  x, = vals
-
-  split_pos = list(axis_names).index(split_name)
-  before_axes = axis_names[:split_pos]
-  after_axes = axis_names[split_pos+1:]
-
-  # Flatten the split_dim
-  split_name_size = psum(1, split_name)
-  before_size = psum(1, before_axes)
-  after_size = psum(1, after_axes)
-  unroll_shape = list(x.shape)
-  unroll_shape[split_axis:split_axis+1] = [before_size, split_name_size, after_size]
-  unroll_x = lax.reshape(x, unroll_shape)
-
-  if before_axes:
-    out_before = all_to_all(unroll_x, before_axes, split_axis, concat_axis=0)
-  else:
-    out_before = _moveaxis(split_axis, 0, unroll_x)
-  out_split = all_to_all(out_before, split_name, split_axis + 1, concat_axis=1)
-  if after_axes:
-    out_after = all_to_all(out_split, after_axes, split_axis + 2, concat_axis=2)
-  else:
-    out_after = _moveaxis(split_axis + 2, 2, out_split)
-
-  # Flatten the concat axes and move them to the right position
-  y = out_after.reshape((np.prod(out_after.shape[:3]), *out_after.shape[3:]))
-  return _moveaxis(0, concat_axis, y)
-
-def _all_to_all_abstract_eval(x, axis_name, split_axis, concat_axis):
+def _all_to_all_abstract_eval(x, axis_name, split_axis, concat_axis, axis_index_groups):
   input_aval = raise_to_shaped(x)
   shape = list(input_aval.shape)
   size = shape.pop(split_axis)
@@ -668,11 +781,10 @@ def _all_to_all_abstract_eval(x, axis_name, split_axis, concat_axis):
 all_to_all_p = core.Primitive('all_to_all')
 all_to_all_p.def_abstract_eval(_all_to_all_abstract_eval)
 xla.parallel_translations[all_to_all_p] = _all_to_all_translation_rule
-ad.deflinear(all_to_all_p, _all_to_all_transpose_rule)
+ad.deflinear2(all_to_all_p, _all_to_all_transpose_rule)
 pxla.multi_host_supported_collectives.add(all_to_all_p)
 batching.primitive_batchers[all_to_all_p] = _all_to_all_batcher
 batching.collective_rules[all_to_all_p] = _all_to_all_batched_collective
-batching.split_axis_rules[all_to_all_p] = _all_to_all_split_axis_rule
 
 
 def _expand(dim, size, index, x):
@@ -681,10 +793,6 @@ def _expand(dim, size, index, x):
   out = lax.full(shape, lax._const(x, 0))
   return lax.dynamic_update_index_in_dim(out, x, index, dim)
 
-def _allgather(x, dim, size, index, axis_name, axis_index_groups=None):
-  outs = tree_util.tree_map(partial(_expand, dim, size, index), x)
-  return psum(outs, axis_name, axis_index_groups=axis_index_groups)
-
 def all_gather(x, axis_name, *, axis_index_groups=None):
   """Gather values of x across all replicas.
 
@@ -737,17 +845,97 @@ def all_gather(x, axis_name, *, axis_index_groups=None):
    [[12. 13. 14. 15.]
     [ 4.  5.  6.  7.]]
   """
+  axis_size = psum(1, axis_name, axis_index_groups=axis_index_groups)
+  # The all_gather primitive doesn't work when omni-staging is disabled.
+  if not config.omnistaging_enabled:
+    return _all_gather_via_psum(x, all_gather_dimension=0, axis_name=axis_name,
+                                axis_index_groups=axis_index_groups, axis_size=axis_size)
 
+  def bind(x):
+    return all_gather_p.bind(x, all_gather_dimension=0, axis_name=axis_name,
+                             axis_index_groups=axis_index_groups, axis_size=axis_size)
+
+  return tree_util.tree_map(bind, x)
+
+def _all_gather_via_psum(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
   index = axis_index(axis_name)
   if axis_index_groups is not None:
     indices = np.array(axis_index_groups).flatten()
     axis_index_to_group_index = indices.argsort() % len(axis_index_groups[0])
     index = lax_numpy.array(axis_index_to_group_index)[index]
+  outs = tree_util.tree_map(partial(_expand, all_gather_dimension, axis_size, index), x)
+  return psum(outs, axis_name, axis_index_groups=axis_index_groups)
 
-  axis_size = psum(1, axis_name, axis_index_groups=axis_index_groups)
+def _all_gather_impl(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+  # Only called when the argument is not mapped.
+  out_shape = list(np.shape(x))
+  out_shape.insert(all_gather_dimension, axis_size)
+  broadcast_dims = [i for i in range(len(out_shape)) if i != all_gather_dimension]
+  return lax.broadcast_in_dim(x, out_shape, broadcast_dims)
 
-  return _allgather(x, 0, axis_size, index, axis_name, axis_index_groups)
+def _all_gather_translation_rule(c, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, axis_env, platform):
+  # TODO(cjfj): Enable this for TPU also?
+  if (platform == 'gpu') and (all_gather_dimension == 0):
+    new_shape = list(c.get_shape(x).dimensions())
+    new_shape.insert(all_gather_dimension, 1)
+    broadcast_dimensions = [i for i in range(len(new_shape)) if i != all_gather_dimension]
+    x = xops.BroadcastInDim(x, new_shape, broadcast_dimensions)
+    replica_groups = _replica_groups(axis_env, axis_name, axis_index_groups)
+    return xops.AllGather(x, all_gather_dimension=all_gather_dimension, shard_count=axis_size,
+                          replica_groups=xc.make_replica_groups(replica_groups))
+  else:
+    lowering = xla.lower_fun(_all_gather_via_psum, multiple_results=False, parallel=True)
+    return lowering(c, x, all_gather_dimension=all_gather_dimension, axis_name=axis_name,
+                    axis_index_groups=axis_index_groups, axis_size=axis_size, axis_env=axis_env, platform=platform)
 
+def _all_gather_abstract_eval(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+  x_aval = raise_to_shaped(x)
+  new_shape = list(x_aval.shape)
+  new_shape.insert(all_gather_dimension, axis_size)
+  return ShapedArray(new_shape, x_aval.dtype)
+
+def _all_gather_transpose_rule(cts, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+  # TODO(cjfj): Add reduce-scatter op to XLA?
+  concat_axis = 0
+  return (lax_numpy.sum(
+      all_to_all(
+          cts,
+          axis_name=axis_name,
+          split_axis=all_gather_dimension,
+          concat_axis=concat_axis,
+          axis_index_groups=axis_index_groups),
+      axis=concat_axis),)
+
+def _all_gather_batcher(vals_in, dims_in, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+  (x,), (d,) = vals_in, dims_in
+  if d <= all_gather_dimension:
+    all_gather_dimension += 1
+  else:
+    d += 1
+  result = all_gather_p.bind(
+      x,
+      all_gather_dimension=all_gather_dimension,
+      axis_name=axis_name,
+      axis_index_groups=axis_index_groups,
+      axis_size=axis_size)
+  return result, d
+
+def _all_gather_batched_collective(frame, vals_in, dims_in, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+  assert axis_index_groups is None, "axis_index_groups not supported in vmap"
+  assert axis_size == frame.size, "axis size doesn't match"
+  assert axis_name == frame.name, "batcher called with wrong axis name"
+  (x,), (d,) = vals_in, dims_in
+  assert d is not batching.not_mapped
+  return _moveaxis(d, all_gather_dimension, x), batching.not_mapped
+
+all_gather_p = core.Primitive('all_gather')
+all_gather_p.def_abstract_eval(_all_gather_abstract_eval)
+all_gather_p.def_impl(_all_gather_impl)
+xla.parallel_translations[all_gather_p] = _all_gather_translation_rule
+ad.deflinear2(all_gather_p, _all_gather_transpose_rule)
+pxla.multi_host_supported_collectives.add(all_gather_p)
+batching.primitive_batchers[all_gather_p] = _all_gather_batcher
+batching.collective_rules[all_gather_p] = _all_gather_batched_collective
 
 def _axis_index_translation_rule(c, *, axis_name, axis_env, platform):
   axis_pos = list(axis_env.names).index(axis_name)
@@ -796,6 +984,78 @@ def _process_axis_index(self, frame):
 batching.BatchTrace.process_axis_index = _process_axis_index  # type: ignore
 
 
+pdot_p = core.Primitive('pdot')
+
+@pdot_p.def_impl
+def _pdot_impl(x, y, *, axis_name, pos_contract, pos_batch):
+  if axis_name: raise NameError(f"unbound axis name: {axis_name[0]}")
+  return lax.dot_general(x, y, [pos_contract, pos_batch])
+
+@pdot_p.def_abstract_eval
+def _pdot_abstract_eval(x, y, *, axis_name, pos_contract, pos_batch):
+  # TODO: avals with names, check inputs are mapped along axis_name, eliminate
+  if not len(set(axis_name)) == len(axis_name): raise ValueError
+  return lax.dot_general_p.abstract_eval(
+      x, y, dimension_numbers=[pos_contract, pos_batch],
+      precision=None, preferred_element_type=None)
+
+def _pdot_vmap_collective_rule(frame, vals_in, dims_in, *, axis_name,
+                               pos_contract, pos_batch):
+  x, y = vals_in
+  x_dim, y_dim = dims_in
+  x_pos_contract, y_pos_contract = pos_contract
+  x_pos_contract = [x_dim] + [d + (d >= x_dim) for d in x_pos_contract]
+  y_pos_contract = [y_dim] + [d + (d >= y_dim) for d in y_pos_contract]
+  x_pos_batch, y_pos_batch = pos_batch
+  x_pos_batch = [d + (d >= x_dim) for d in x_pos_batch]
+  y_pos_batch = [d + (d >= y_dim) for d in y_pos_batch]
+  remaining_axis_names = tuple(n for n in axis_name if n != frame.name)
+  out = pdot_p.bind(x, y, axis_name=remaining_axis_names,
+                    pos_contract=[x_pos_contract, y_pos_contract],
+                    pos_batch=[x_pos_batch, y_pos_batch])
+  return out, None
+batching.collective_rules[pdot_p] = _pdot_vmap_collective_rule
+
+def _pdot_vmap_batching_rule(vals_in, dims_in, *, axis_name, pos_contract,
+                             pos_batch):
+  x, y = vals_in
+  (pos_contract, pos_batch), result_batch_dim = lax._dot_general_batch_dim_nums(
+      (x.ndim, y.ndim), dims_in, [pos_contract, pos_batch])
+  out = pdot_p.bind(x, y, axis_name=axis_name, pos_contract=pos_contract,
+                    pos_batch=pos_batch)
+  return out, result_batch_dim
+batching.primitive_batchers[pdot_p] = _pdot_vmap_batching_rule
+
+def _pdot_translation_rule(c, x, y, *, axis_name, pos_contract, pos_batch,
+                           axis_env, platform):
+  local_out = lax._dot_general_translation_rule(
+      c, x, y, dimension_numbers=[pos_contract, pos_batch], precision=None,
+      preferred_element_type=None)
+  if axis_name:
+    out_tup = xla.parallel_translations[psum_p](
+        c, local_out, axis_name=axis_name, axis_index_groups=None,
+        axis_env=axis_env, platform=platform)
+    out, = xla.xla_destructure(c, out_tup)
+  else:
+    out = local_out
+  return out
+xla.parallel_translations[pdot_p] = _pdot_translation_rule
+
+def _pdot_transpose_lhs(g, y, *, axis_name, pos_contract, pos_batch):
+  # TODO: avals with names, call pbroadcast with axis_name
+  return lax._dot_general_transpose_lhs(
+      g, y, dimension_numbers=[pos_contract, pos_batch], precision=None,
+      preferred_element_type=None)
+def _pdot_transpose_rhs(g, x, *, axis_name, pos_contract, pos_batch):
+  # TODO: avals with names, call pbroadcast with axis_name
+  return lax._dot_general_transpose_rhs(
+      g, x, dimension_numbers=[pos_contract, pos_batch], precision=None,
+      preferred_element_type=None)
+ad.defbilinear(pdot_p, _pdot_transpose_lhs, _pdot_transpose_rhs)
+
+pxla.multi_host_supported_collectives.add(pdot_p)
+
+
 @config.register_omnistaging_disabler
 def omnistaging_disabler() -> None:
   global axis_index
