@@ -12,12 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from typing import Any, Tuple
+import operator
+from typing import Any, Tuple, Union
 
 from .tree_util import (tree_flatten, tree_unflatten, tree_multimap, _replace_nones,
-                        tree_structure)
+                        tree_structure, treedef_children, treedef_is_leaf)
 from . import linear_util as lu
-from .util import safe_map, WrapHashably, Hashable
+from ._src.util import safe_map, WrapHashably, Hashable
 from .core import unit
 
 from ._src import traceback_util
@@ -25,6 +26,19 @@ traceback_util.register_exclusion(__file__)
 
 map = safe_map
 
+def _ensure_index(x: Any) -> Union[int, Tuple[int, ...]]:
+  """Ensure x is either an index or a tuple of indices."""
+  try:
+    return operator.index(x)
+  except TypeError:
+    return tuple(map(operator.index, x))
+
+def _ensure_index_tuple(x: Any) -> Tuple[int, ...]:
+  """Convert x to a tuple of indices."""
+  try:
+    return (operator.index(x),)
+  except TypeError:
+    return tuple(map(operator.index, x))
 
 @lu.transformation_with_aux
 def flatten_fun(in_tree, *args_flat):
@@ -64,10 +78,7 @@ def flatten_fun_nokwargs2(in_tree, *args_flat):
 
 
 def argnums_partial(f, dyn_argnums, args):
-  if isinstance(dyn_argnums, int):
-    dyn_argnums = (dyn_argnums,)
-  else:
-    dyn_argnums = tuple(dyn_argnums)
+  dyn_argnums = _ensure_index_tuple(dyn_argnums)
   fixed_args = tuple([unit if i in dyn_argnums else wrap_hashably(arg)
                       for i, arg in enumerate(args)])
   dyn_args = tuple(args[i] for i in dyn_argnums)
@@ -155,7 +166,7 @@ def wrap_hashably(arg):
   else:
     return Hashable(arg)
 
-def flatten_axes(name, treedef, axis_tree):
+def flatten_axes(name, treedef, axis_tree, *, kws=False):
   # given an axis spec tree axis_tree (a pytree with integers and Nones at the
   # leaves, i.e. the Nones are to be considered leaves) that is a tree prefix of
   # the given treedef, build a complete axis spec tree with the same structure
@@ -168,6 +179,12 @@ def flatten_axes(name, treedef, axis_tree):
   try:
     tree_multimap(add_leaves, _replace_nones(proxy, axis_tree), dummy)
   except ValueError:
+    if kws:
+      # if keyword arguments are included in the tree, we make adapt the error
+      # message only to be about the positional arguments
+      treedef, leaf = treedef_children(treedef)
+      assert treedef_is_leaf(leaf)
+      axis_tree, _ = axis_tree
     raise ValueError(f"{name} specification must be a tree prefix of the "
                      f"corresponding value, got specification {axis_tree} "
                      f"for value tree {treedef}.") from None
