@@ -35,6 +35,10 @@ def bool_env(varname: str, default: bool) -> bool:
   else:
     raise ValueError("invalid truth value %r for environment %r" % (val, varname))
 
+def int_env(varname: str, default: int) -> int:
+  """Read an environment variable and interpret it as an integer."""
+  return int(os.getenv(varname, default))
+
 
 class Config:
   def __init__(self):
@@ -160,3 +164,17 @@ flags.DEFINE_bool(
     bool_env('JAX_OMNISTAGING', True),
     help='Enable staging based on dynamic context rather than data dependence.'
 )
+
+flags.DEFINE_integer(
+    'jax_tracer_error_num_traceback_frames',
+    int_env('JAX_TRACER_ERROR_NUM_TRACEBACK_FRAMES', 5),
+    help='Set the number of stack frames in JAX tracer error messages.'
+)
+
+flags.DEFINE_bool(
+    'jax_check_tracer_leaks',
+    bool_env('JAX_CHECK_TRACER_LEAKS', False),
+    help=('Turn on checking for leaked tracers as soon as a trace completes. '
+          'Enabling leak checking may have performance impacts: some caching '
+          'is disabled, and other overheads may be added.'),
+)
