@@ -24,7 +24,7 @@ import threading
 import types
 from typing import (Any, Callable, ClassVar, Dict, Generator,
                     Iterator, List, NamedTuple, Optional, Sequence, Set, Tuple,
-                    Type, Union, cast)
+                    Type, Union, cast, Iterable, Hashable)
 
 import numpy as np
 
@@ -33,22 +33,20 @@ from .config import FLAGS, config
 from . import linear_util as lu
 
 from jax._src import source_info_util
-from .util import (safe_zip, safe_map, partial, curry, prod, partialmethod,
-                   tuple_insert, tuple_delete)
-from .pprint_util import pp, vcat, PrettyPrint
+from ._src.util import (safe_zip, safe_map, partial, curry, prod, partialmethod,
+                   tuple_insert, tuple_delete, as_hashable_function,
+                   HashableFunction)
+from ._src.pprint_util import pp, vcat, PrettyPrint
 
 from ._src import traceback_util
 traceback_util.register_exclusion(__file__)
 
-# TODO(dougalm): compilation cache breaks the leak detector. Consisder solving.
-check_leaks = False
-
-# Disables internal invariant checks
-skip_checks = not FLAGS.jax_enable_checks  # not __debug__  # google doesn't use -O
+# TODO(mattjj): move this into debug_state
+skip_checks = not FLAGS.jax_enable_checks
 
 @contextmanager
 def skipping_checks():
-  """Context manager for temporarily disabling checks."""
+  """Context manager for temporarily disabling internal checks."""
   global skip_checks
   old_value, skip_checks = skip_checks, True
   try:
@@ -56,6 +54,20 @@ def skipping_checks():
   finally:
     skip_checks = old_value
 
+@contextmanager
+def checking_leaks():
+  """Context manager for temporarily enabling tracer leak checks."""
+  old_value, debug_state.check_leaks = debug_state.check_leaks, True
+  try:
+    yield
+  finally:
+    debug_state.check_leaks = old_value
+
+class DebugState(threading.local):
+  def __init__(self):
+    self.check_leaks = FLAGS.jax_check_tracer_leaks
+debug_state = DebugState()
+
 zip = safe_zip
 map = safe_map
 
@@ -312,6 +324,13 @@ def extract_call_jaxpr(
     return (params["call_jaxpr"], new_params)
 
 
+def traverse_jaxpr_params(f, params):
+  """Applies f to each jaxpr parameter and returns a tuple of returned values."""
+  return tuple(f(param if type(param) is Jaxpr else param.jaxpr)
+               for param in params.values()
+               if type(param) in (Jaxpr, ClosedJaxpr))
+
+
 def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
   def read(v):
     if type(v) is Literal:
@@ -333,8 +352,15 @@ def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
       subfuns = [lu.wrap_init(partial(eval_jaxpr, call_jaxpr, ()))]
     else:
       subfuns = []
+    if eqn.primitive.map_primitive:
+      out_axes_thunk = HashableFunction(lambda: params['out_axes'],
+                                        closure=params['out_axes'])
+      bind_params = dict(params, out_axes_thunk=out_axes_thunk)
+      del bind_params['out_axes']
+    else:
+      bind_params = params
     with source_info_util.user_context(eqn.source_info):
-      ans = eqn.primitive.bind(*(subfuns + in_vals), **params)
+      ans = eqn.primitive.bind(*(subfuns + in_vals), **bind_params)
     if eqn.primitive.multiple_results:
       map(write, eqn.outvars, ans)
     else:
@@ -369,19 +395,19 @@ class Trace:
       elif val._trace.sublevel < sublevel:
         return self.sublift(val)
       else:
-        raise escaped_tracer_error("Can't lift sublevels {} to {}"
-                                   .format(val._trace.sublevel, sublevel))
+        raise escaped_tracer_error(
+            val, f"Can't lift sublevels {val._trace.sublevel} to {sublevel}")
     elif val._trace.level < level:
       if val._trace.sublevel > sublevel:
-        raise escaped_tracer_error("Incompatible sublevel: {}, {}"
-                                   .format(val._trace, (level, sublevel)))
+        raise escaped_tracer_error(
+            val, f"Incompatible sublevel: {val._trace}, {(level, sublevel)}")
       return self.lift(val)
     elif val._trace.level > level:
-      raise escaped_tracer_error("Can't lift level {} to {}"
-                                 .format(val, self))
+      raise escaped_tracer_error(
+          val, f"Can't lift level {val} to {self}")
     else:  # val._trace.level == self.level:
-      raise escaped_tracer_error("Different traces at same level: {}, {}"
-                                 .format(val, self))
+      raise escaped_tracer_error(
+          val, f"Different traces at same level: {val}, {self}")
 
   def pure(self, val):
     raise NotImplementedError("must override")
@@ -419,20 +445,42 @@ class Trace:
            "to handle custom_vjp primitives")
     raise NotImplementedError(msg)
 
-def escaped_tracer_error(detail=None):
+def escaped_tracer_error(tracer, detail=None):
+  num_frames = FLAGS.jax_tracer_error_num_traceback_frames
   msg = ("Encountered an unexpected tracer. Perhaps this tracer escaped "
          "through global state from a previously traced function.\n"
          "The functions being transformed should not save traced values to "
          "global state.")
   if detail:
     msg += " Detail: {}.".format(detail)
+  try:
+    line_info = tracer._line_info
+  except AttributeError:
+    pass
+  else:
+    msg += ('\nThe tracer that caused this error was created on line '
+            f'{source_info_util.summarize(line_info)}.\n')
+    if num_frames > 0:
+      msg += (f'When the tracer was created, the final {num_frames} stack '
+              'frames (most recent last) excluding JAX-internal frames were:\n'
+              f'{source_info_util.summarize(line_info, num_frames=num_frames)}')
+  try:
+    fun_source_info = tracer._trace.main.source_info
+  except AttributeError:
+    pass
+  else:
+    msg += ('\nThe function being traced when the tracer leaked was '
+            f'{fun_source_info}.')
+  msg += ('\nTo catch the leak earlier, try setting the environment variable '
+          'JAX_CHECK_TRACER_LEAKS or using the `jax.checking_leaks` context '
+          'manager.')
   return UnexpectedTracerError(msg)
 
 class UnexpectedTracerError(Exception): pass
 
 class Tracer:
   __array_priority__ = 1000
-  __slots__ = ['_trace', '__weakref__']
+  __slots__ = ['_trace', '__weakref__', '_line_info']
 
   def __array__(self, *args, **kw):
     msg = ("The numpy.ndarray conversion method __array__() was called on "
@@ -648,6 +696,7 @@ class TraceStack:
 
 class Sublevel(int): pass
 AxisEnvFrame = namedtuple('AxisEnvFrame', ['name', 'size', 'main_trace'])
+AxisName = Hashable
 
 class TraceState:
   trace_stack: TraceStack
@@ -707,16 +756,15 @@ def new_main(trace_type: Type[Trace],
   try:
     yield main
   finally:
-    thread_local_state.trace_state.trace_stack.pop()
+    stack.pop()
     if dynamic:
       stack.dynamic = prev_dynamic
 
-  if check_leaks:
+  if debug_state.check_leaks:
     t = ref(main)
     del main
     if t() is not None:
-      print(thread_local_state.trace_state.trace_stack)
-      raise Exception('Leaked trace {}'.format(t()))
+      raise Exception(f'Leaked trace {t()}')
 
 @contextmanager
 def new_base_main(trace_type: Type[Trace]) -> Generator[MainTrace, None, None]:
@@ -731,6 +779,12 @@ def new_base_main(trace_type: Type[Trace]) -> Generator[MainTrace, None, None]:
     stack.dynamic = prev_dynamic
     stack.stack[0] = prev_base
 
+  if debug_state.check_leaks:
+    t = ref(main)
+    del main
+    if t() is not None:
+      raise Exception('Leaked trace {}'.format(t()))
+
 @contextmanager
 def eval_context():
   with new_base_main(EvalTrace):
@@ -745,11 +799,12 @@ def new_sublevel() -> Generator[None, None, None]:
   finally:
     thread_local_state.trace_state.substack.pop()
 
-  if check_leaks:
-    t = ref(sublevel)
-    del sublevel
-    if t() is not None:
-      raise Exception('Leaked sublevel {}'.format(t()))
+  # TODO(mattjj): to check sublevel leaks, we need to make Sublevel weakref-able
+  # if debug_state.check_leaks:
+  #   t = ref(sublevel)
+  #   del sublevel
+  #   if t() is not None:
+  #     raise Exception('Leaked sublevel {}'.format(t()))
 
 def maybe_new_sublevel(trace):
   # dynamic traces run the WrappedFun, so we raise the sublevel for them
@@ -763,8 +818,13 @@ def full_lower(val):
     return val
 
 def find_top_trace(xs) -> Trace:
-  top_main = max((x._trace.main for x in xs if isinstance(x, Tracer)),
-                 default=None, key=attrgetter('level'))
+  top_tracer = max((x for x in xs if isinstance(x, Tracer)),
+                    default=None, key=attrgetter('_trace.level'))
+  if top_tracer is not None:
+    top_tracer._assert_live()
+    top_main = top_tracer._trace.main  # type: Optional[MainTrace]
+  else:
+    top_main = None
   dynamic = thread_local_state.trace_state.trace_stack.dynamic
   top_main = (dynamic if top_main is None or dynamic.level > top_main.level
               else top_main)
@@ -779,7 +839,7 @@ class AbstractValue:
   _num_buffers: int = 1  # number of buffers used to represent the value.
 
   def at_least_vspace(self):
-    return self
+    raise NotImplementedError("must override")
 
   def __repr__(self):
     try:
@@ -801,6 +861,7 @@ bot = Bot()
 class AbstractUnit(AbstractValue):
   # TODO(jakevdp): make it possible to set zero buffers
   # _num_buffers = 0
+  def at_least_vspace(self): return self
   def join(self, other):
     if not skip_checks:
       assert other is abstract_unit, other
@@ -1093,8 +1154,9 @@ class AbstractToken(AbstractValue):
     else:
       assert False, f"Cannot join {self} with {other}"
   def str_short(self): return 'Tok'
+  def at_least_vspace(self): return self
 
-abstract_token = AbstractToken()
+abstract_token: AbstractToken = AbstractToken()
 
 
 def raise_to_shaped(aval: AbstractValue, weak_type=None):
@@ -1150,12 +1212,19 @@ def apply_todos(todos, outs):
     outs = map(full_lower, todos_list.pop()(outs))
   return outs
 
+class _IgnoreElemList(list):
+  """Compares equal to all other _ignore_elem_lists."""
+  def __hash__(self): return 0
+  def __eq__(self, other):
+    return type(other) is _IgnoreElemList
+
 @lu.transformation_with_aux
 def process_env_traces(primitive: Union['CallPrimitive', 'MapPrimitive'],
-                       level: int, params_tuple: tuple, *args):
+                       level: int, params_tuple: tuple, out_axes_transforms, *args):
   outs = yield args, {}
   params = dict(params_tuple)
   todo = []
+  assert not out_axes_transforms
   while True:
     tracers = [x for x in outs if isinstance(x, Tracer)
                and (level is None or x._trace.level > level)]
@@ -1166,15 +1235,32 @@ def process_env_traces(primitive: Union['CallPrimitive', 'MapPrimitive'],
     trace = ans._trace.main.with_cur_sublevel()
     outs = map(trace.full_raise, outs)
     outs, cur_todo = primitive.post_process(trace, outs, params)
+    if isinstance(primitive, MapPrimitive):
+      cur_todo, out_axes_transform = cur_todo
+      out_axes_transforms.append(out_axes_transform)
     todo.append(cur_todo)
   yield outs, tuple(todo)  # Ensure the aux output is immutable
 
 def call_bind(primitive: Union['CallPrimitive', 'MapPrimitive'],
               fun, *args, **params):
+  out_axes_transforms = _IgnoreElemList()
+  if primitive.map_primitive:
+    out_axes_thunk = params['out_axes_thunk']
+    # The new thunk depends deterministically on the old thunk and the wrapped function.
+    # Any caching already has to include the wrapped function as part of the key, so we
+    # only use the previous thunk for equality checks.
+    @as_hashable_function(closure=out_axes_thunk)
+    def new_out_axes_thunk():
+      out_axes = out_axes_thunk()
+      for t in out_axes_transforms:
+        out_axes = t(out_axes)
+      return out_axes
+    params = dict(params, out_axes_thunk=new_out_axes_thunk)
   params_tuple = tuple(params.items())
   top_trace = find_top_trace(args)
   fun, env_trace_todo = process_env_traces(
-      fun, primitive, top_trace and top_trace.level, params_tuple)
+      fun, primitive, top_trace and top_trace.level,
+      params_tuple, out_axes_transforms)
   tracers = map(top_trace.full_raise, args)
   with maybe_new_sublevel(top_trace):
     outs = primitive.process(top_trace, fun, tracers, params)
@@ -1222,7 +1308,7 @@ class MapPrimitive(Primitive):
     return trace.post_process_map(self, out_tracers, params)
 
 @contextmanager
-def extend_axis_env(axis_name, size: int, tag: Any):
+def extend_axis_env(axis_name: AxisName, size: int, tag: Any):
   frame = AxisEnvFrame(axis_name, size, tag)
   thread_local_state.trace_state.axis_env.append(frame)
   try:
@@ -1230,6 +1316,16 @@ def extend_axis_env(axis_name, size: int, tag: Any):
   finally:
     thread_local_state.trace_state.axis_env.pop()
 
+@contextmanager
+def extend_axis_env_nd(axes: Iterable[Tuple[AxisName, int]]):
+  frames = [AxisEnvFrame(axis_name, size, None) for axis_name, size in axes]
+  thread_local_state.trace_state.axis_env.extend(frames)
+  try:
+    yield
+  finally:
+    for _ in frames:
+      thread_local_state.trace_state.axis_env.pop()
+
 
 # When a mapped function is given no axis name, we generate a name object based
 # on the id of the function object. Collisions aren't important because this
@@ -1257,16 +1353,11 @@ def axis_frame(axis_name):
   for frame in reversed(frames):
     if frame.name == axis_name:
       return frame
-
-  named_axis = [
-      frame.name
-      for frame in reversed(frames)
-      if not isinstance(frame.name, _TempAxisName)
-  ]
+  named_axes = [frame.name for frame in reversed(frames)
+                if not isinstance(frame.name, _TempAxisName)]
   raise NameError(
       f'unbound axis name: {axis_name}. The following axis names (e.g. defined '
-      'by pmap) are available to collective operations:'
-      f'{named_axis}')
+      f'by pmap) are available to collective operations: {named_axes}')
 
 
 # ------------------- Jaxpr checking -------------------
@@ -1421,6 +1512,9 @@ def check_map(prim, in_avals, params):
   typecheck_assert("in_axes" in params,
                    f"Map primitive {prim} missing 'in_axes' parameter")
   in_axes = params["in_axes"]
+  typecheck_assert("out_axes" in params,
+                   f"Map primitive {prim} missing 'out_axes' parameter")
+  out_axes = params["out_axes"]
 
   binder_avals = [unmapped_aval(axis_size, in_axis, v.aval)
                   if in_axis is not None else v.aval
@@ -1436,8 +1530,8 @@ def check_map(prim, in_avals, params):
   _check_jaxpr(call_jaxpr, mapped_avals)
 
   mapped_out_avals = [v.aval for v in call_jaxpr.outvars]
-  # Assuming out_axes == 0
-  out_avals = [unmapped_aval(axis_size, 0, aval) for aval in mapped_out_avals]
+  out_avals = [unmapped_aval(axis_size, out_axis, aval) if out_axis is not None else aval
+               for aval, out_axis in zip(mapped_out_avals, out_axes)]
   return out_avals
 
 
@@ -1604,7 +1698,7 @@ def omnistaging_disabler() -> None:
     finally:
       thread_local_state.trace_state.trace_stack.pop(bottom)
 
-    if check_leaks:
+    if debug_state.check_leaks:
       t = ref(main)
       del main
       if t() is not None:
@@ -1637,12 +1731,26 @@ def omnistaging_disabler() -> None:
 
   def call_bind(primitive: Union['CallPrimitive', 'MapPrimitive'],
                 fun: lu.WrappedFun, *args, **params):
+    out_axes_transforms = _IgnoreElemList()
+    if primitive.map_primitive:
+      out_axes_thunk = params['out_axes_thunk']
+      # The new thunk depends deterministically on the old thunk and the wrapped function.
+      # Any caching already has to include the wrapped function as part of the key, so we
+      # only use the previous thunk for equality checks.
+      @as_hashable_function(closure=out_axes_thunk)
+      def new_out_axes_thunk():
+        out_axes = out_axes_thunk()
+        for t in out_axes_transforms:
+          out_axes = t(out_axes)
+        return out_axes
+      params = dict(params, out_axes_thunk=new_out_axes_thunk)
     params_tuple = tuple(params.items())
     top_trace = find_top_trace(args)
     level = (thread_local_state.trace_state.trace_stack.next_level(True)
             if top_trace is None else top_trace.level)
     params_tuple = tuple(params.items())
-    fun, env_trace_todo = process_env_traces(fun, primitive, level, params_tuple)
+    fun, env_trace_todo = process_env_traces(
+        fun, primitive, level, params_tuple, out_axes_transforms)
     if top_trace is None:
       with new_sublevel():
         outs = primitive.impl(fun, *args, **params)
