@@ -19,12 +19,13 @@ import operator as op
 from typing import Callable, Sequence, Tuple, Any
 
 from . import core
+from . import dtypes
 from . import linear_util as lu
 from .tree_util import (tree_flatten, tree_unflatten, tree_map, tree_multimap,
-                        register_pytree_node_class)
-from .util import safe_zip, safe_map, split_list
+                        treedef_is_leaf, register_pytree_node_class)
+from ._src.util import cache, safe_zip, safe_map, split_list
 from .api_util import flatten_fun_nokwargs, argnums_partial, wrap_hashably
-from .abstract_arrays import raise_to_shaped
+from .core import raise_to_shaped
 from .ad_util import Zero, zeros_like_aval, stop_gradient_p
 from .interpreters import partial_eval as pe
 from .interpreters import ad
@@ -88,7 +89,7 @@ class custom_jvp:
   :py:func:`~jax.custom_jvp.defjvp` for defining a *single* custom JVP rule for
   all the function's inputs, and for convenience
   :py:func:`~jax.custom_jvp.defjvps`, which wraps
-  :py:func:`~jax.custom_jvp.defjvp`, and allows you to provide seperate
+  :py:func:`~jax.custom_jvp.defjvp`, and allows you to provide separate
   definitions for the partial derivatives of the function w.r.t. each of its
   arguments.
 
@@ -270,9 +271,9 @@ class CustomJVPCallPrimitive(core.CallPrimitive):
     args = map(core.full_lower, args)
     top_trace = core.find_top_trace(args)
     fun, env_trace_todo1 = core.process_env_traces(
-        fun, self, top_trace and top_trace.level, ())
+        fun, self, top_trace and top_trace.level, (), None)
     jvp, env_trace_todo2 = core.process_env_traces(
-        jvp, self, top_trace and top_trace.level, ())
+        jvp, self, top_trace and top_trace.level, (), None)
     tracers = map(top_trace.full_raise, args)  # type: ignore
     with core.maybe_new_sublevel(top_trace):
       outs = top_trace.process_custom_jvp_call(self, fun, jvp, tracers)  # type: ignore
@@ -569,9 +570,9 @@ class CustomVJPCallPrimitive(core.CallPrimitive):
     args = map(core.full_lower, args)
     top_trace = core.find_top_trace(args)
     fun, env_trace_todo1 = core.process_env_traces(
-        fun, self, top_trace and top_trace.level, ())
+        fun, self, top_trace and top_trace.level, (), None)
     fwd, env_trace_todo2 = core.process_env_traces(
-        fwd, self, top_trace and top_trace.level, ())
+        fwd, self, top_trace and top_trace.level, (), None)
     tracers = map(top_trace.full_raise, args)  # type: ignore
     with core.maybe_new_sublevel(top_trace):
       outs = top_trace.process_custom_vjp_call(self, fun, fwd, bwd, tracers,
@@ -626,13 +627,14 @@ def _custom_vjp_call_jaxpr_vmap(
     args, in_dims, axis_name, *, fun_jaxpr: core.ClosedJaxpr,
     fwd_jaxpr_thunk: Callable[[], Tuple[core.Jaxpr, Sequence[Any]]],
     bwd: lu.WrappedFun, out_trees: Callable, num_consts: int):
-  size, = {x.shape[d] for x, d in zip(args, in_dims) if d is not not_mapped}
+  axis_size, = {x.shape[d] for x, d in zip(args, in_dims) if d is not not_mapped}
   args = [batching.moveaxis(x, d, 0) if d is not not_mapped and d != 0
           else x for x, d in zip(args, in_dims)]
 
   in_batched = [d is not not_mapped for d in in_dims]
   _, args_batched = split_list(in_batched, [num_consts])
-  batched_fun_jaxpr, out_batched = batching.batch_jaxpr(fun_jaxpr, size, in_batched, False, axis_name)
+  batched_fun_jaxpr, out_batched = batching.batch_jaxpr(
+      fun_jaxpr, axis_size, in_batched, False, axis_name)
   out_dims1 = [0 if b else not_mapped for b in out_batched]
   out_dims2 = []
 
@@ -640,15 +642,14 @@ def _custom_vjp_call_jaxpr_vmap(
   def batched_fwd_jaxpr_thunk():
     fwd_jaxpr = core.ClosedJaxpr(*fwd_jaxpr_thunk())  # consts can be tracers
     batched_fwd_jaxpr, out_batched = batching.batch_jaxpr(
-        fwd_jaxpr, size, args_batched, False, axis_name)
+        fwd_jaxpr, axis_size, args_batched, False, axis_name)
     out_dims2.append([0 if b else not_mapped for b in out_batched])
     return batched_fwd_jaxpr.jaxpr, batched_fwd_jaxpr.consts
 
   fwd_args_batched = [0 if b else not_mapped for b in args_batched]
   fwd_out_dims = lambda: out_dims2[0]
-  # TODO(mattjj,apaszke): Support collectives in custom_vjp?
-  batched_bwd = batching.batch_fun(bwd, fwd_out_dims, fwd_args_batched,
-                                   axis_name='__unused_axis_name', sum_match=True)
+  batched_bwd = batching.batch(bwd, axis_name, axis_size, fwd_out_dims,
+                               fwd_args_batched)
 
   batched_outs = custom_vjp_call_jaxpr_p.bind(
       *args, fun_jaxpr=batched_fun_jaxpr,
@@ -681,9 +682,9 @@ def omnistaging_disabler() -> None:
     args = map(core.full_lower, args)
     top_trace = core.find_top_trace(args)
     fun, env_trace_todo1 = core.process_env_traces(
-        fun, self, top_trace and top_trace.level, ())
+        fun, self, top_trace and top_trace.level, (), None)
     jvp, env_trace_todo2 = core.process_env_traces(
-        jvp, self, top_trace and top_trace.level, ())
+        jvp, self, top_trace and top_trace.level, (), None)
     if top_trace is None:
       with core.new_sublevel():
         outs = self.impl(fun, jvp, *args)
@@ -817,7 +818,10 @@ def custom_gradient(fun):
     cts_flat, out_tree_ = tree_flatten((cts,))
     if out_tree != out_tree_: raise TypeError(f'{out_tree}\n!=\n{out_tree_}')
     cts_out = core.eval_jaxpr(jaxpr, consts, *cts_flat)
-    return tree_unflatten(in_tree, cts_out)
+    cts_out = tree_unflatten(in_tree, cts_out)
+    if treedef_is_leaf(in_tree):
+      cts_out = (cts_out,)
+    return cts_out
 
   wrapped_fun.defvjp(fwd, bwd)
   return wrapped_fun
@@ -837,3 +841,111 @@ class Residuals:
   def tree_unflatten(cls, aux, consts):
     jaxpr, in_tree, out_tree = aux
     return cls(jaxpr, in_tree, out_tree, consts)
+
+
+def closure_convert(fun, *example_args):
+  """Closure conversion utility, for use with higher-order custom derivatives.
+
+  To define custom derivatives such as with ``jax.custom_vjp(f)``, the target
+  function ``f`` must take, as formal arguments, all values involved in
+  differentiation. If ``f`` is a higher-order function, in that it accepts as an
+  argument a Python function ``g``, then values stored away in ``g``'s closure
+  will not be visible to the custom derivative rules, and attempts at AD
+  involving these values will fail. One way around this is to convert the
+  closure by extracting these values, and to pass them as explicit formal
+  arguments across the custom derivative boundary. This utility carries out that
+  conversion. More precisely, it closure-converts the function ``fun``
+  specialized to the types of the arguments given in ``example_args``.
+
+  When we refer here to "values in the closure" of ``fun``, we do not mean the
+  values that are captured by Python directly when ``fun`` is defined (e.g. the
+  Python objects in ``fun.__closure__``, if the attribute exists). Rather, we
+  mean values encountered during the execution of ``fun`` on ``example_args``
+  that determine its output. This may include, for instance, arrays captured
+  transitively in Python closures, i.e. in the Python closure of functions
+  called by ``fun``, the closures of the functions that they call, and so forth.
+
+  The function ``fun`` must be a pure function.
+
+  Example usage::
+
+    def minimize(objective_fn, x0):
+      converted_fn, aux_args = closure_convert(objective_fn, x0)
+      return _minimize(converted_fn, x0, *aux_args)
+
+    @partial(custom_vjp, nondiff_argnums=(0,))
+    def _minimize(objective_fn, x0, *args):
+      z = objective_fn(x0, *args)
+      # ... find minimizer x_opt ...
+      return x_opt
+
+    def fwd(objective_fn, x0, *args):
+      y = _minimize(objective_fn, x0, *args)
+      return y, (y, args)
+
+    def rev(objective_fn, res, g):
+      y, args = res
+      y_bar = g
+      # ... custom reverse-mode AD ...
+      return x0_bar, *args_bars
+
+    _minimize.defvjp(fwd, rev)
+
+  Args:
+    fun: Python callable to be converted. Must be a pure function.
+    example_args: Arrays, scalars, or (nested) standard Python
+      containers (tuples, lists, dicts, namedtuples, i.e., pytrees)
+      thereof, used to determine the types of the formal arguments to
+      ``fun``. This type-specialized form of ``fun`` is the function
+      that will be closure converted.
+
+  """
+  flat_args, in_tree = tree_flatten(example_args)
+  in_avals = tuple(map(abstractify, flat_args))
+  if core.debug_state.check_leaks:
+    return _closure_convert_for_avals.__wrapped__(fun, in_tree, in_avals)
+  else:
+    return _closure_convert_for_avals(fun, in_tree, in_avals)
+
+@cache()
+def _closure_convert_for_avals(fun, in_tree, in_avals):
+  if config.omnistaging_enabled:
+    wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
+    jaxpr, out_pvals, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals)
+  else:
+    in_pvals = [pe.PartialVal.unknown(aval) for aval in in_avals]
+    wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
+    with core.initial_style_staging():  # type: ignore
+      jaxpr, out_pvals, consts = pe.trace_to_jaxpr(
+        wrapped_fun, in_pvals, instantiate=True, stage_out=False)  # type: ignore
+  out_tree = out_tree()
+
+  # We only want to closure convert for constants with respect to which we're
+  # differentiating. As a proxy for that, we hoist consts with float dtype.
+  # TODO(mattjj): revise this approach
+  from .numpy import inexact
+  is_float = lambda c: dtypes.issubdtype(dtypes.dtype(c), inexact)
+  (closure_consts, hoisted_consts), merge = partition_list(is_float, consts)
+  num_consts = len(hoisted_consts)
+
+  def converted_fun(*args_hconsts):
+    num_args = len(args_hconsts) - num_consts
+    args, hoisted_consts = split_list(args_hconsts, [num_args])
+    consts = merge(closure_consts, hoisted_consts)
+    all_args, in_tree2 = tree_flatten(tuple(args))
+    assert in_tree == in_tree2
+    out_flat = core.eval_jaxpr(jaxpr, consts, *all_args)
+    return tree_unflatten(out_tree, out_flat)
+
+  return converted_fun, hoisted_consts
+
+def partition_list(choice, lst):
+  out = [], []
+  which = [out[choice(elt)].append(elt) or choice(elt) for elt in lst]
+  def merge(l1, l2):
+    i1, i2 = iter(l1), iter(l2)
+    return [next(i2 if snd else i1) for snd in which]
+  return out, merge
+
+def abstractify(x):
+  return core.raise_to_shaped(core.get_aval(x))
