@@ -23,10 +23,11 @@
 from distutils.util import strtobool
 import functools
 import os
+from typing import Dict
 
 import numpy as np
 
-from . import util
+from ._src import util
 from .config import flags
 from .lib import xla_client
 
@@ -42,21 +43,6 @@ flags.DEFINE_bool('jax_enable_x64',
 bfloat16 = xla_client.bfloat16
 _bfloat16_dtype = np.dtype(bfloat16)
 
-class _bfloat16_finfo(object):
-  bits = 16
-  eps = bfloat16(float.fromhex("0x1p-7"))
-  epsneg = bfloat16(float.fromhex("0x1p-8"))
-  machep = -7
-  negep = -8
-  max = bfloat16(float.fromhex("0x1.FEp127"))
-  min = -max
-  nexp = 8
-  nmant = 7
-  iexp = nexp
-  precision = 2
-  resolution = 10 ** -2
-  tiny = bfloat16(float.fromhex("0x1p-126"))
-
 # Default types.
 
 bool_ = np.bool_
@@ -128,14 +114,51 @@ def coerce_to_array(x):
 
 iinfo = np.iinfo
 
-def finfo(dtype):
-  # Since NumPy doesn't consider bfloat16 a floating-point type, we have to
-  # provide an alternative implementation of finfo that does so.
-  if ((isinstance(dtype, str) and dtype == "bfloat16") or
-      np.result_type(dtype) == _bfloat16_dtype):
-    return _bfloat16_finfo
-  else:
-    return np.finfo(dtype)
+class finfo(np.finfo):
+  __doc__ = np.finfo.__doc__
+  _finfo_cache: Dict[np.dtype, np.finfo] = {}
+  @staticmethod
+  def _bfloat16_finfo():
+    def float_to_str(f):
+      return "%12.4e" % float(f)
+
+    bfloat16 = _bfloat16_dtype.type
+    tiny = float.fromhex("0x1p-126")
+    resolution = 0.01
+    eps = float.fromhex("0x1p-7")
+    epsneg = float.fromhex("0x1p-8")
+    max = float.fromhex("0x1.FEp127")
+
+    obj = object.__new__(np.finfo)
+    obj.dtype = _bfloat16_dtype
+    obj.bits = 16
+    obj.eps = bfloat16(eps)
+    obj.epsneg = bfloat16(epsneg)
+    obj.machep = -7
+    obj.negep = -8
+    obj.max = bfloat16(max)
+    obj.min = bfloat16(-max)
+    obj.nexp = 8
+    obj.nmant = 7
+    obj.iexp = obj.nexp
+    obj.precision = 2
+    obj.resolution = bfloat16(resolution)
+    obj.tiny = bfloat16(tiny)
+    obj.machar = None  # np.core.getlimits.MachArLike does not support bfloat16.
+
+    obj._str_tiny = float_to_str(tiny)
+    obj._str_max = float_to_str(max)
+    obj._str_epsneg = float_to_str(epsneg)
+    obj._str_eps = float_to_str(eps)
+    obj._str_resolution = float_to_str(resolution)
+    return obj
+
+  def __new__(cls, dtype):
+    if isinstance(dtype, str) and dtype == 'bfloat16' or dtype == _bfloat16_dtype:
+      if _bfloat16_dtype not in cls._finfo_cache:
+        cls._finfo_cache[_bfloat16_dtype] = cls._bfloat16_finfo()
+      return cls._finfo_cache[_bfloat16_dtype]
+    return super().__new__(cls, dtype)
 
 def _issubclass(a, b):
   """Determines if ``a`` is a subclass of ``b``.
@@ -165,6 +188,17 @@ def issubdtype(a, b):
 can_cast = np.can_cast
 issubsctype = np.issubsctype
 
+# Return the type holding the real part of the input type
+def dtype_real(typ):
+  if np.issubdtype(typ, np.complexfloating):
+    if typ == np.dtype('complex64'):
+      return np.dtype('float32')
+    elif typ == np.dtype('complex128'):
+      return np.dtype('float64')
+    else:
+      raise TypeError("Unknown complex floating type {}".format(typ))
+  else:
+    return typ
 
 # Enumeration of all valid JAX types in order.
 _weak_types = [int, float, complex]
@@ -226,7 +260,7 @@ def _make_lattice_upper_bounds():
   return upper_bounds
 _lattice_upper_bounds = _make_lattice_upper_bounds()
 
-@functools.lru_cache(512)
+@functools.lru_cache(512)  # don't use util.memoize because there is no X64 dependence.
 def _least_upper_bound(*nodes):
   # This function computes the least upper bound of a set of nodes N within a partially
   # ordered set defined by the lattice generated above.
