@@ -28,7 +28,7 @@ from jax.tree_util import tree_flatten, tree_unflatten
 from jax.api_util import flatten_fun_nokwargs
 from jax import ad_util, core, lax, xla
 from jax._src.lax import lax as lax_internal
-from jax.util import unzip2, wrap_name
+from jax._src.util import unzip2, wrap_name
 import jax.numpy as jnp
 import jax.linear_util as lu
 
@@ -244,11 +244,11 @@ _def_inequality(lax.le_p, operator.le)
 _def_inequality(lax.eq_p, operator.eq)
 _def_inequality(lax.ne_p, operator.ne)
 
-def _convert_element_type(operand, new_dtype, old_dtype):
+def _convert_element_type(operand, new_dtype):
   head, tail = operand
-  head = lax.convert_element_type_p.bind(head, new_dtype=new_dtype, old_dtype=old_dtype)
+  head = lax.convert_element_type_p.bind(head, new_dtype=new_dtype)
   if tail is not None:
-    tail = lax.convert_element_type_p.bind(tail, new_dtype=new_dtype, old_dtype=old_dtype)
+    tail = lax.convert_element_type_p.bind(tail, new_dtype=new_dtype)
   if jnp.issubdtype(new_dtype, jnp.floating):
     if tail is None:
       tail = jnp.zeros_like(head)
