@@ -21,11 +21,12 @@ from . import partial_eval as pe
 from ..config import config
 from .. import core
 from ..dtypes import dtype, float0
-from ..core import Trace, Tracer, get_aval, call_p, Primitive, Literal
+from ..core import (Trace, Tracer, get_aval, call_p, Primitive, Literal,
+                    raise_to_shaped)
 from ..ad_util import (add_jaxvals, add_jaxvals_p, zeros_like_jaxval, zeros_like_aval,
                        zeros_like_p, Zero)
-from ..abstract_arrays import raise_to_shaped
-from ..util import unzip2, safe_map, safe_zip, partial, split_list, wrap_name, moveaxis
+from .._src.util import (unzip2, safe_map, safe_zip, partial, split_list,
+                         wrap_name, as_hashable_function)
 from ..tree_util import register_pytree_node
 from .. import linear_util as lu
 from ..api_util import flatten_fun, flatten_fun_nokwargs
@@ -163,20 +164,20 @@ def backward_pass(jaxpr: core.Jaxpr, consts, primals_in, cotangents_in):
   if all(type(ct) is Zero for ct in cotangents_in):
     return map(lambda v: Zero(v.aval), jaxpr.invars)
 
-  def write_cotangent(v, ct):
+  def write_cotangent(prim, v, ct):
     # assert v not in primal_env
+    assert ct is not Zero, (prim, v.aval)  # check for an old harmless type error
     if ct is None or type(v) is Literal:
       return
     if type(ct) is Zero:
       # FIXME: This triggers a lot of failures!
-      # assert v.aval == ct.aval, (v.aval, ct.aval)
-      return
-    if ct is Zero: # FIXME: This should never happen, but seems to be necessary
+      # assert v.aval == ct.aval, (prim, v.aval, ct.aval)
       return
     ct_env[v] = add_tangents(ct_env[v], ct) if v in ct_env else ct
     if not core.skip_checks:
       ct_aval = core.get_aval(ct_env[v])
-      assert v.aval.strip_weak_type() == core.lattice_join(v.aval, ct_aval).strip_weak_type(), (v.aval, ct_aval)
+      joined_aval = core.lattice_join(v.aval, ct_aval).strip_weak_type()
+      assert v.aval.strip_weak_type() == joined_aval, (prim, v.aval, ct_aval)
 
   def read_cotangent(v):
     return ct_env.get(v, Zero(v.aval))
@@ -208,7 +209,7 @@ def backward_pass(jaxpr: core.Jaxpr, consts, primals_in, cotangents_in):
     seen_vars |= read_set
 
   ct_env: Dict[Any, Any] = {}
-  map(write_cotangent, jaxpr.outvars, cotangents_in)
+  map(partial(write_cotangent, 'outvars'), jaxpr.outvars, cotangents_in)
   for eqn, to_drop in zip(jaxpr.eqns[::-1], drop_cts[::-1]):
     # FIXME: Some invars correspond to tangents
     invals = map(read_primal, eqn.invars)
@@ -227,7 +228,7 @@ def backward_pass(jaxpr: core.Jaxpr, consts, primals_in, cotangents_in):
                                                          **eqn.params)
     cts_out = [Zero(v.aval) for v in eqn.invars] if cts_out is Zero else cts_out
     # FIXME: Some invars correspond to primals!
-    map(write_cotangent, eqn.invars, cts_out)
+    map(partial(write_cotangent, eqn.primitive), eqn.invars, cts_out)
     for var in to_drop:
       ct_env.pop(var, None)  # NB: Constant cotangents might be missing
 
@@ -256,6 +257,12 @@ def get_primitive_transpose(p):
         "Transpose rule (for reverse-mode differentiation) for '{}' "
         "not implemented".format(p)) from err
 
+@lu.transformation_with_aux
+def nonzero_tangent_outputs(*args, **kwargs):
+  results = (_, tangents_out) = yield args, kwargs
+  yield results, [type(r) is not Zero for r in tangents_out]
+
+
 class JVPTrace(Trace):
 
   def pure(self, val):
@@ -285,14 +292,27 @@ class JVPTrace(Trace):
     assert call_primitive.multiple_results
     primals, tangents = unzip2((t.primal, t.tangent) for t in tracers)
     nonzero_tangents, tangent_tree_def = tree_flatten(tangents)
-    f_jvp, out_tree_def = traceable(jvp_subtrace(f, self.main),
-                                    len(primals), tangent_tree_def)
     nz_tangents = [type(t) is not Zero for t in tangents]
     params = dict(params, name=wrap_name(params['name'], 'jvp'))
+    f_jvp = jvp_subtrace(f, self.main)
     if isinstance(call_primitive, core.MapPrimitive):
       in_axes = params['in_axes']
       tangent_in_axes = [ax for ax, nz in zip(in_axes, nz_tangents) if nz]
-      params = dict(params, in_axes=(*in_axes, *tangent_in_axes))
+      out_axes_thunk = params['out_axes_thunk']
+      f_jvp, nz_tangents_out = nonzero_tangent_outputs(f_jvp)
+      # The new thunk depends deterministically on the old thunk and the wrapped function.
+      # Any caching already has to include the wrapped function as part of the key, so we
+      # only use the previous thunk for equality checks.
+      # NOTE: This assumes that the output tangents being zero is a deterministic
+      #       function of which input tangents were zero.
+      @as_hashable_function(closure=(tuple(nz_tangents), out_axes_thunk))
+      def new_out_axes_thunk():
+        out_axes = out_axes_thunk()
+        return (*out_axes, *(ax for ax, nz in zip(out_axes, nz_tangents_out()) if nz))
+      params = dict(params,
+                    in_axes=(*in_axes, *tangent_in_axes),
+                    out_axes_thunk=new_out_axes_thunk)
+    f_jvp, out_tree_def = traceable(f_jvp, len(primals), tangent_tree_def)
     update_params = call_param_updaters.get(call_primitive)
     new_params = update_params(params, nz_tangents) if update_params else params
     result = call_primitive.bind(f_jvp, *primals, *nonzero_tangents, **new_params)
@@ -302,16 +322,22 @@ class JVPTrace(Trace):
   def post_process_call(self, call_primitive, out_tracers, params):
     primals, tangents = unzip2((t.primal, t.tangent) for t in out_tracers)
     out, treedef = tree_flatten((primals, tangents))
+    tangents_nz = [type(t) is not Zero for t in tangents]
     del primals, tangents
     main = self.main
     def todo(x):
       primals, tangents = tree_unflatten(treedef, x)
       trace = JVPTrace(main, core.cur_sublevel())
       return map(partial(JVPTracer, trace), primals, tangents)
+    if call_primitive.map_primitive:
+      def out_axes_transform(out_axes):
+        return (*out_axes, *(ax for ax, nz in zip(out_axes, tangents_nz) if nz))
+      todo = (todo, out_axes_transform)
     return out, todo
 
   # The only difference between process_map and process_call is that
-  # the `in_axes` param must be updated; that's handled in process_call.
+  # the `in_axes` and `out_axes_thunk` params must be updated;
+  # that's handled in process_call.
   process_map = process_call
   post_process_map = post_process_call
 
@@ -461,7 +487,7 @@ def defbilinear_broadcasting(bcast, prim, lhs_rule, rhs_rule):
   rhs_jvp = lambda g, x, y, **kwargs: prim.bind(x, bcast(g, x), **kwargs)
   defjvp(prim, lhs_jvp, rhs_jvp)
   primitive_transposes[prim] = partial(bilinear_transpose, lhs_rule, rhs_rule)
-defbilinear = partial(defbilinear_broadcasting, lambda g, x: g)
+defbilinear: Callable = partial(defbilinear_broadcasting, lambda g, x: g)
 
 def bilinear_transpose(lhs_rule, rhs_rule, cotangent, x, y, **kwargs):
   assert is_undefined_primal(x) ^ is_undefined_primal(y)
@@ -484,8 +510,8 @@ def zero_jvp(primitive, primals, tangents, **params):
   return r, Zero.from_value(r)
 
 
-deflinear(zeros_like_p, lambda t: [Zero.from_value(t)])
-deflinear(add_jaxvals_p, lambda t: (t, t))
+deflinear2(zeros_like_p, lambda t, _: [Zero.from_value(t)])
+deflinear2(add_jaxvals_p, lambda t, *args: (t, t))
 
 def instantiate_zeros(tangent):
   if type(tangent) is Zero:
@@ -495,7 +521,7 @@ def instantiate_zeros(tangent):
   else:
     return tangent
 
-# This function seems similar to instantaite_zeros, but it is sometimes used
+# This function seems similar to instantiate_zeros, but it is sometimes used
 # to instantiate zero abstract units with a different aval
 def instantiate_zeros_aval(aval, tangent):
   if type(tangent) is Zero:
@@ -558,18 +584,34 @@ def remat_transpose(params, call_jaxpr, primals_in, cotangents_in, cotangent_in_
   return tree_unflatten(out_tree(), flat_cotangents_out)
 primitive_transposes[pe.remat_call_p] = remat_transpose
 
+@lu.transformation_with_aux
+def nonzero_outputs(*args, **kwargs):
+  results = yield args, kwargs
+  yield results, [type(r) is not Zero for r in results]
+
 
 def map_transpose(primitive, params, call_jaxpr, args, ct, _):
   all_args, in_tree_def = tree_flatten(((), args, ct))  # empty consts
   fun = lu.hashable_partial(lu.wrap_init(backward_pass), call_jaxpr)
+  fun, nz_arg_cts = nonzero_outputs(fun)
   fun, out_tree = flatten_fun_nokwargs(fun, in_tree_def)
-  # Preserve axis for primal arguments, skip tangents (represented as undefined primals),
-  # we only support out_axes=0 for now, so hard-code that for each cotangent.
-  new_in_axes = (*[axis for axis, x in zip(params['in_axes'], args)
+  # Preserve axis for primal arguments, skip tangents (represented as undefined primals).
+  in_axes, out_axes = params['in_axes'], params['out_axes']
+  new_in_axes = (*[axis for axis, x in zip(in_axes, args)
                    if not is_undefined_primal(x)],
-                 *[0 for x in ct if type(x) is not Zero])
+                 *[axis for axis, x in zip(out_axes, ct)
+                   if type(x) is not Zero])
+  # The interim strategy we use below (until avals-with-names) only works
+  # when all outputs are mapped.
+  assert all(out_axis is not None for out_axis in out_axes), out_axes
+  # NOTE: This assumes that the output cotangents being zero is a deterministic
+  #       function of which input cotangents were zero.
+  @as_hashable_function(closure=(in_axes, tuple(type(c) is Zero for c in ct)))
+  def out_axes_thunk():
+    return tuple(axis or 0 for axis, nz in zip(in_axes, nz_arg_cts()) if nz)
   new_params = dict(params, name=wrap_name(params['name'], 'transpose'),
-                    in_axes=new_in_axes)
+                    in_axes=new_in_axes, out_axes_thunk=out_axes_thunk)
+  del new_params['out_axes']
   update_params = call_transpose_param_updaters.get(primitive)
   if update_params:
     new_params = update_params(new_params, map(is_undefined_primal, args),
@@ -577,7 +619,6 @@ def map_transpose(primitive, params, call_jaxpr, args, ct, _):
   out_flat = primitive.bind(fun, *all_args, **new_params)
   arg_cts = tree_unflatten(out_tree(), out_flat)
 
-  in_axes = params['in_axes']
   # The freevars are being fanned out (not mapped). During transpose the
   # dual of fan-out is fan-in-sum. We apply it to the unmapped invars.
   assert len(in_axes) == len(arg_cts)
@@ -585,7 +626,7 @@ def map_transpose(primitive, params, call_jaxpr, args, ct, _):
     return (zero if in_axis is None else
             Zero(core.unmapped_aval(params['axis_size'], in_axis, zero.aval)))
   arg_cts = (unmap_zero(arg_ct, in_axis) if type(arg_ct) is Zero else
-             moveaxis(arg_ct, 0, in_axis) if in_axis is not None else
+             arg_ct if in_axis is not None else
              arg_ct.sum(0)
              for arg_ct, in_axis in zip(arg_cts, in_axes))
   return tuple(arg_cts)
