@@ -18,62 +18,64 @@ from typing import Any, Callable, Dict, Optional, Tuple, Union
 import jax
 from ..config import config
 from .. import core
-from ..core import Trace, Tracer
-from ..abstract_arrays import ShapedArray, raise_to_shaped
+from ..core import ShapedArray, raise_to_shaped, Trace, Tracer
 from ..ad_util import add_jaxvals, add_jaxvals_p, zeros_like_jaxval, zeros_like_p
 from .. import linear_util as lu
-from ..util import (unzip2, partial, safe_map, wrap_name, split_list,
-                    canonicalize_axis, moveaxis)
+from .._src.util import (unzip2, partial, safe_map, wrap_name, split_list,
+                         canonicalize_axis, moveaxis, as_hashable_function)
 from . import xla
 from . import partial_eval as pe
 
 map = safe_map
 
+def batch(fun: lu.WrappedFun, axis_name, axis_size, in_dims, out_dim_dests,
+          ) -> lu.WrappedFun:
+  # anlogue of `jvp` in ad.py
+  fun, out_dims_thunk = batch_subtrace(fun)
+  return _match_axes(batchfun(fun, axis_name, axis_size, in_dims),
+                     axis_size, out_dims_thunk, out_dim_dests)
 
-def batch(fun: lu.WrappedFun, in_vals, in_dims, out_dim_dests, axis_name):
-  # executes a batched version of `fun` following out_dim_dests
-  batched_fun = batch_fun(fun, in_dims, out_dim_dests, axis_name=axis_name)
-  return batched_fun.call_wrapped(*in_vals)
+@lu.transformation
+def batchfun(axis_name, axis_size, in_dims, *in_vals):
+  # analogue of `jvpfun` in ad.py
+  in_dims = in_dims() if callable(in_dims) else in_dims
+  in_dims = [canonicalize_axis(ax, np.ndim(x)) if isinstance(ax, int)
+             and not isinstance(core.get_aval(x), core.AbstractUnit)  # non-omnistaging
+             else ax for x, ax in zip(in_vals, in_dims)]
+  with core.new_main(BatchTrace, axis_name=axis_name) as main:
+    with core.extend_axis_env(axis_name, axis_size, main):
+      out_vals = yield (main, in_dims, *in_vals), {}
+      del main
+  yield out_vals
 
 @lu.transformation_with_aux
-def batch_subtrace(main, in_dims, *in_vals, **params):
+def batch_subtrace(main, in_dims, *in_vals):
+  # analogue of `jvp_subtrace` in ad.py
   trace = main.with_cur_sublevel()
   in_tracers = [BatchTracer(trace, val, dim) if dim is not None else val
                 for val, dim in zip(in_vals, in_dims)]
-  outs = yield in_tracers, params
+  outs = yield in_tracers, {}
   out_tracers = map(trace.full_raise, outs)
   out_vals, out_dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
   yield out_vals, out_dims
 
-
-def batch_fun(fun : lu.WrappedFun, in_dims, out_dim_dests, axis_name,
-              sum_match=False):
-  # transformation version of batch, which doesn't call the function
-  fun, out_dims = batch_subtrace(fun)
-  return _batch_fun(fun, axis_name, sum_match, in_dims, out_dims, out_dim_dests)
-
 @lu.transformation
-def _batch_fun(axis_name, sum_match, in_dims, out_dims_thunk, out_dim_dests,
-               *in_vals, **params):
-  in_dims = in_dims() if callable(in_dims) else in_dims
-  in_dims = [
-    canonicalize_axis(dim, np.ndim(val)) if isinstance(dim, int) else dim
-    for val, dim in zip(in_vals, in_dims)]
-  size, = {x.shape[d] for x, d in zip(in_vals, in_dims) if d is not not_mapped}
-  with core.new_main(BatchTrace, axis_name=axis_name) as main:
-    with core.extend_axis_env(axis_name, size, main):
-      out_vals = yield (main, in_dims,) + in_vals, params
-    del main
+def _match_axes(axis_size, out_dims_thunk, out_dim_dests, *in_vals):
+  out_vals = yield in_vals, {}
   out_dim_dests = out_dim_dests() if callable(out_dim_dests) else out_dim_dests
   out_dims = out_dims_thunk()
   for od, od_dest in zip(out_dims, out_dim_dests):
-    if od is not None and not isinstance(od_dest, int) and not sum_match:
+    if od is not None and not isinstance(od_dest, int):
       msg = f"vmap has mapped output but out_axes is {od_dest}"
       raise ValueError(msg)
-  out_vals = map(partial(matchaxis, size, sum_match=sum_match), out_dims, out_dim_dests, out_vals)
-  yield out_vals
+  yield map(partial(matchaxis, axis_size), out_dims, out_dim_dests, out_vals)
 
-def batch_fun2(fun : lu.WrappedFun, in_dims):
+
+# These next two functions, `batch_fun2` and `_batch_fun2`, are deprecated; the
+# former is only called from `custom_transforms`, which itself is deprecated.
+# TODO(mattjj): delete these along with custom_transforms
+
+def batch_fun2(fun: lu.WrappedFun, in_dims):
   # like `batch_fun` but returns output batch dims (so no out_dim_dests)
   fun, out_dims = batch_subtrace(fun)
   return _batch_fun2(fun, in_dims), out_dims
@@ -140,25 +142,13 @@ class BatchTrace(Trace):
     vals_in, dims_in = unzip2((t.val, t.batch_dim) for t in tracers)
     if all(bdim is not_mapped for bdim in dims_in):
       return primitive.bind(*vals_in, **params)
-    elif config.omnistaging_enabled and primitive in collective_rules:
-      axis_names = params['axis_name']
-      if not isinstance(axis_names, (tuple, list)):
-        axis_names = (axis_names,)
-      for i, axis_name in enumerate(axis_names):
-        frame = core.axis_frame(axis_name)
-        if frame.main_trace is not self.main:
-          continue
-        # We run the split_axis rule with tracers, which is supposed to never
-        # mix this axis name with another one. We will handle any invocations
-        # of collectives over the vmapped axis in a recursive call from a tracer.
-        if len(axis_names) > 1:
-          return split_axis(primitive, axis_name, tracers, params)
-        vals_out, dims_out = collective_rules[primitive](vals_in, dims_in, frame.size, **params)
-        results = map(partial(BatchTracer, self), vals_out, dims_out)
-        return results if primitive.multiple_results else results[0]
-    # TODO(mattjj,phawkins): if no rule implemented, could vmap-via-map here
-    batched_primitive = get_primitive_batcher(primitive, self.axis_name)
-    val_out, dim_out = batched_primitive(vals_in, dims_in, **params)
+    if (primitive in collective_rules and
+          _main_trace_for_axis_names(self.main, params['axis_name'])):
+      frame = core.axis_frame(self.axis_name)
+      val_out, dim_out = collective_rules[primitive](frame, vals_in, dims_in, **params)
+    else:
+      batched_primitive = get_primitive_batcher(primitive, self.axis_name)
+      val_out, dim_out = batched_primitive(vals_in, dims_in, **params)
     if primitive.multiple_results:
       return map(partial(BatchTracer, self), val_out, dim_out)
     else:
@@ -200,26 +190,43 @@ class BatchTrace(Trace):
       # When both d and in_axis are defined then:
       # - If `d <= in_axis`, we have to move the `in_axis` one dimension further;
       # - If `d >  in_axis`, we have to decrement `d` (as `in_axis` will get removed).
-      def fmap(x, f): return None if x is None else f(x)
-      new_in_axes = tuple(fmap(in_axis,
-                               lambda ax: ax + (1 if d is not None and d <= in_axis else 0))
-                          for d, in_axis in zip(dims, params['in_axes']))
-      new_dims = tuple(fmap(d,
-                            lambda dv: dv - (1 if in_axis is not None and in_axis < d else 0))
-                       for d, in_axis in zip(dims, params['in_axes']))
+      def both_mapped(in_out_axis, d):
+        return in_out_axis is not None and d is not not_mapped
+      new_in_axes = tuple(
+        in_axis + 1 if both_mapped(in_axis, d) and d <= in_axis else in_axis
+        for d, in_axis in zip(dims, params['in_axes']))
+      new_dims = tuple(
+        d - 1 if both_mapped(in_axis, d) and in_axis < d else d
+        for d, in_axis in zip(dims, params['in_axes']))
       f, dims_out = batch_subtrace(f, self.main, new_dims)
-      vals_out = map_primitive.bind(f, *vals, **dict(params, in_axes=new_in_axes))
-      # We assume that out_axes are all 0, so we increment d
-      dims_out = tuple(d + 1 if d is not not_mapped else d for d in dims_out())
+      out_axes_thunk = params['out_axes_thunk']
+      # NOTE: This assumes that the choice of the dimensions over which outputs
+      #       are batched is entirely dependent on the function and not e.g. on the
+      #       data or its shapes.
+      @as_hashable_function(closure=out_axes_thunk)
+      def new_out_axes_thunk():
+        return tuple(out_axis + 1 if both_mapped(out_axis, d) and d < out_axis else out_axis
+                     for out_axis, d in zip(out_axes_thunk(), dims_out()))
+      new_params = dict(params, in_axes=new_in_axes, out_axes_thunk=new_out_axes_thunk)
+      vals_out = map_primitive.bind(f, *vals, **new_params)
+      dims_out = (d + 1 if both_mapped(out_axis, d) and out_axis <= d else d
+                  for d, out_axis in zip(dims_out(), out_axes_thunk()))
       return [BatchTracer(self, v, d) for v, d in zip(vals_out, dims_out)]
 
   def post_process_map(self, call_primitive, out_tracers, params):
     vals, dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
     main = self.main
+    def both_mapped(in_out_axis, d):
+      return in_out_axis is not None and d is not not_mapped
     def todo(vals):
       trace = main.with_cur_sublevel()
-      return [BatchTracer(trace, v, d + 1 if d is not not_mapped else d)
-              for v, d in zip(vals, dims)]
+      return [BatchTracer(trace, v, d + 1 if both_mapped(out_axis, d) and out_axis <= d else d)
+              for v, d, out_axis in zip(vals, dims, params['out_axes_thunk']())]
+    if call_primitive.map_primitive:
+      def out_axes_transform(out_axes):
+        return tuple(out_axis + 1 if both_mapped(out_axis, d) and d < out_axis else out_axis
+                     for out_axis, d in zip(out_axes, dims))
+      todo = (todo, out_axes_transform)
     return vals, todo
 
   def process_custom_jvp_call(self, prim, fun, jvp, tracers):
@@ -243,17 +250,42 @@ class BatchTrace(Trace):
 
   def process_custom_vjp_call(self, prim, fun, fwd, bwd, tracers, *, out_trees):
     in_vals, in_dims = unzip2((t.val, t.batch_dim) for t in tracers)
+    axis_size, = {x.shape[d] for x, d in zip(in_vals, in_dims)
+                  if d is not not_mapped}
     fun, out_dims1 = batch_subtrace(fun, self.main, in_dims)
     fwd, out_dims2 = batch_subtrace(fwd, self.main, in_dims)
-    # TODO(mattjj,apaszke): support collectives in custom_vjp?
-    bwd = batch_fun(bwd, out_dims2, in_dims,
-                    axis_name='__unused_axis_name', sum_match=True)
+    bwd = batch_custom_vjp_bwd(bwd, self.axis_name, axis_size,
+                               out_dims2, in_dims)
     out_vals = prim.bind(fun, fwd, bwd, *in_vals, out_trees=out_trees)
     fst, out_dims = lu.merge_linear_aux(out_dims1, out_dims2)
     if not fst:
       out_dims = out_dims[-len(out_vals) % len(out_dims):]
     return [BatchTracer(self, v, d) for v, d in zip(out_vals, out_dims)]
 
+  post_process_custom_vjp_call = post_process_custom_jvp_call
+
+def _main_trace_for_axis_names(main_trace: core.MainTrace,
+                               axis_name: Union[core.AxisName, Tuple[core.AxisName, ...]]
+                               ) -> bool:
+  # This function exists to identify whether a main trace corresponds to any of
+  # the axis names used by a primitive. Axis names alone aren't enough because
+  # axis names can shadow, so we use the main trace as a tag.
+  if not isinstance(axis_name, (list, tuple)):
+    axis_name = (axis_name,)
+  return any(main_trace is core.axis_frame(n).main_trace for n in axis_name)
+
+def batch_custom_vjp_bwd(bwd, axis_name, axis_size, in_dims, out_dim_dests):
+  bwd, out_dims_thunk = batch_subtrace(bwd)
+  return _match_axes_and_sum(batchfun(bwd, axis_name, axis_size, in_dims),
+                             axis_size, out_dims_thunk, out_dim_dests)
+
+@lu.transformation
+def _match_axes_and_sum(axis_size, out_dims_thunk, out_dim_dests, *in_vals):
+  # this is like _match_axes, but we do reduce-sums as needed
+  out_vals = yield in_vals, {}
+  yield map(partial(matchaxis, axis_size, sum_match=True),
+            out_dims_thunk(), out_dim_dests, out_vals)
+
 
 ### primitives
 
@@ -382,40 +414,35 @@ def bdim_at_front(x, bdim, size):
     return moveaxis(x, bdim, 0)
 
 
-def _promote_aval_rank(sz, aval):
-  if aval is core.abstract_unit:
-    return core.abstract_unit
-  else:
-    return ShapedArray((sz,) + aval.shape, aval.dtype)
-
-def batch_jaxpr(closed_jaxpr, size, batched, instantiate, axis_name):
+def batch_jaxpr(closed_jaxpr, axis_size, in_batched, instantiate, axis_name):
   f = lu.wrap_init(core.jaxpr_as_fun(closed_jaxpr))
-  f, batched_out = batched_traceable(f, size, batched, instantiate, axis_name)
-  avals_in = [_promote_aval_rank(size, a) if b else a
-              for a, b in zip(closed_jaxpr.in_avals, batched)]
+  f, out_batched = batch_subtrace_instantiate(f, instantiate, axis_size)
+  f = batchfun(f, axis_name, axis_size, [0 if b else None for b in in_batched])
+  avals_in = [core.unmapped_aval(axis_size, 0, aval) if b else aval
+              for aval, b in zip(closed_jaxpr.in_avals, in_batched)]
   jaxpr_out, _, consts = pe.trace_to_jaxpr_dynamic(f, avals_in)
-  return core.ClosedJaxpr(jaxpr_out, consts), batched_out()
+  return core.ClosedJaxpr(jaxpr_out, consts), out_batched()
 
 @lu.transformation_with_aux
-def batched_traceable(size, batched, instantiate, axis_name, *vals):
-  in_dims = [0 if b else None for b in batched]
-  with core.new_main(BatchTrace, axis_name=axis_name) as main:
-    with core.extend_axis_env(axis_name, size, main):
-      trace = main.with_cur_sublevel()
-      ans = yield map(partial(BatchTracer, trace), vals, in_dims), {}
-      out_tracers = map(trace.full_raise, ans)
-      out_vals, out_dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
-      del main, out_tracers
+def batch_subtrace_instantiate(instantiate, axis_size, main, in_dims, *in_vals):
+  # this is like `batch_subtrace` but we take an extra `instantiate` arg
+  # analogue of `jvp_subtrace` in ad.py
+  trace = main.with_cur_sublevel()
+  in_tracers = [BatchTracer(trace, val, dim) if dim is not None else val
+                for val, dim in zip(in_vals, in_dims)]
+  outs = yield in_tracers, {}
+  out_tracers = map(trace.full_raise, outs)
+  out_vals, out_dims = unzip2((t.val, t.batch_dim) for t in out_tracers)
+
   if type(instantiate) is bool:
     instantiate = [instantiate] * len(out_vals)
   out_vals = [moveaxis(x, d, 0) if d is not not_mapped and d != 0
-              else broadcast(x, size, 0) if d is not_mapped and inst else x
+              else broadcast(x, axis_size, 0) if d is not_mapped and inst else x
               for x, d, inst in zip(out_vals, out_dims, instantiate)]
   out_batched = [d is not not_mapped or inst
                  for d, inst in zip(out_dims, instantiate)]
   yield out_vals, out_batched
 
-
 @lu.transformation_with_aux
 def batch_custom_jvp_subtrace(main, in_dims, *in_vals):
   size, = {x.shape[d] for x, d in zip(in_vals, in_dims) if d is not not_mapped}
@@ -447,21 +474,16 @@ def _merge_bdims(x, y):
 def omnistaging_disabler() -> None:
   global batch_jaxpr
 
-  def batch_jaxpr(jaxpr, size, batched, instantiate, axis_name):
+  def batch_jaxpr(jaxpr, axis_size, in_batched, instantiate, axis_name):
     f = lu.wrap_init(core.jaxpr_as_fun(jaxpr))
-    f, batched_out = batched_traceable(f, size, batched, instantiate, axis_name)
-    avals_in = [_promote_aval_rank(size, a) if b else a
-                for a, b in zip(jaxpr.in_avals, batched)]
+    f, out_batched = batch_subtrace_instantiate(f, instantiate, axis_size)
+    f = batchfun(f, axis_name, axis_size, [0 if b else None for b in in_batched])
+    avals_in = [core.unmapped_aval(axis_size, 0, aval) if b else aval
+                for aval, b in zip(jaxpr.in_avals, in_batched)]
     in_pvals = [pe.PartialVal.unknown(aval) for aval in avals_in]
     jaxpr_out, pvals_out, consts_out = pe.trace_to_jaxpr(f, in_pvals, instantiate=True)
     avals_out, _ = unzip2(pvals_out)
-    return core.ClosedJaxpr(jaxpr_out, consts_out), batched_out()
+    return core.ClosedJaxpr(jaxpr_out, consts_out), out_batched()
 
 
-# collective_rules can assume that the collective is only carried out throughout
-# the vmapped axis (i.e. no tuples in axis_name).
 collective_rules: Dict[core.Primitive, Callable] = {}
-split_axis_rules: Dict[core.Primitive, Callable] = {}
-
-def split_axis(primitive, split_name, args, params):
-  return split_axis_rules[primitive](split_name, args, params)
