@@ -24,7 +24,7 @@ from . import partial_eval as pe
 from ..core import raise_to_shaped, get_aval, Literal, Jaxpr
 from ..api_util import flatten_fun_nokwargs
 from ..tree_util import tree_flatten, tree_unflatten, register_pytree_node
-from ..util import safe_map, safe_zip, split_list
+from .._src.util import safe_map, safe_zip, split_list
 from .. import custom_derivatives
 from ..config import config
 
