@@ -13,10 +13,11 @@
 # limitations under the License.
 
 
-from collections import defaultdict, deque, namedtuple
+from collections import defaultdict, deque
 import itertools as it
 import operator as op
-from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Type, Tuple
+from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Type,
+                    Tuple, Union, NamedTuple)
 from warnings import warn
 
 from absl import logging
@@ -29,12 +30,11 @@ from .. import dtypes
 from .. import lazy
 from .. import linear_util as lu
 from jax._src import source_info_util
-from ..abstract_arrays import (ConcreteArray, ShapedArray, AbstractToken,
-                               make_shaped_array, array_types, raise_to_shaped,
-                               abstract_token)
-from ..core import Literal, pp_eqn_compact
-from ..pprint_util import pp
-from ..util import (partial, partialmethod, cache, prod, unzip2,
+from ..abstract_arrays import (make_shaped_array, array_types)
+from ..core import (ConcreteArray, ShapedArray, AbstractToken,
+                    Literal, pp_eqn_compact, raise_to_shaped, abstract_token)
+from jax._src.pprint_util import pp
+from .._src.util import (partial, partialmethod, cache, prod, unzip2,
                     extend_name_stack, wrap_name, safe_zip, safe_map)
 from ..lib import xla_bridge as xb
 from ..lib import xla_client as xc
@@ -62,6 +62,9 @@ FLAGS = flags.FLAGS
 flags.DEFINE_bool('jax_debug_nans',
                   bool_env('JAX_DEBUG_NANS', False),
                   'Add nan checks to every operation.')
+flags.DEFINE_bool('jax_debug_infs',
+                  bool_env('JAX_DEBUG_INFS', False),
+                  'Add inf checks to every operation.')
 flags.DEFINE_bool('jax_log_compiles',
                   bool_env('JAX_LOG_COMPILES', False),
                   'Print a message each time a `jit` computation is compiled.')
@@ -275,7 +278,7 @@ def xla_primitive_callable(prim, *arg_specs: Tuple[core.AbstractValue,
         f"compiling a primitive computation `{prim}` that requires {nreps} "
         f"replicas, but only {xb.device_count(backend)} XLA devices are "
         f"available on backend {backend.platform}.")
-  built_c = primitive_computation(prim, AxisEnv(nreps, (), (), None), backend,
+  built_c = primitive_computation(prim, AxisEnv(nreps, (), ()), backend,
                                   tuple_args, *avals, **params)
   options = xb.get_compile_options(
       num_replicas=nreps,
@@ -320,6 +323,9 @@ def primitive_computation(prim, axis_env, backend, tuple_args, *avals, **params)
   elif prim in translations:
     rule = translations[prim]
     ans = rule(c, *xla_args, **params)
+  elif prim in translations_with_avals:
+    rule = translations_with_avals[prim]
+    ans = rule(c, avals, xla_args, params)
   elif prim in initial_style_translations:
     rule = initial_style_translations[prim]
     ans = rule(c, axis_env, extend_name_stack(prim.name), avals, backend,
@@ -329,7 +335,7 @@ def primitive_computation(prim, axis_env, backend, tuple_args, *avals, **params)
   assert isinstance(ans, xe.XlaOp)
   c.clear_op_metadata()
   try:
-    return c.build()
+    return c.build(ans)
   except RuntimeError as e:
     msg = (" ".join(map(str, e.args)) + "\n"
            "This is a bug in JAX's shape-checking rules; please report it!\n"
@@ -337,7 +343,7 @@ def primitive_computation(prim, axis_env, backend, tuple_args, *avals, **params)
     raise RuntimeError(msg) from e
 
 def primitive_subcomputation(prim, *avals, **params):
-  axis_env = AxisEnv(1, (), (), None)
+  axis_env = AxisEnv(1, (), ())
   return primitive_computation(prim, axis_env, None, False, *avals, **params)
 
 def backend_compile(backend, built_c, options):
@@ -349,7 +355,7 @@ def _execute_compiled_primitive(prim, compiled, result_handler, *args):
   device, = compiled.local_devices()
   input_bufs = list(it.chain.from_iterable(device_put(x, device) for x in args if x is not token))
   out_bufs = compiled.execute(input_bufs)
-  if FLAGS.jax_debug_nans: check_nans(prim, out_bufs)
+  check_special(prim, out_bufs)
   return result_handler(*out_bufs)
 
 def _execute_replicated_primitive(prim, compiled, result_handler, *args):
@@ -360,17 +366,18 @@ def _execute_replicated_primitive(prim, compiled, result_handler, *args):
   return result_handler(*out_bufs)
 
 
-def check_nans(prim, bufs):
-  for buf in bufs:
-    # TODO(jblespiau): We can simply use buf.xla_shape() when version 0.1.58 is
-    # the default.
-    _check_nans(prim.name, getattr(buf, "xla_shape", buf.shape)(), buf)
+def check_special(prim, bufs):
+  if FLAGS.jax_debug_infs or FLAGS.jax_debug_nans:
+    for buf in bufs:
+      _check_special(prim.name, buf.xla_shape(), buf)
 
-def _check_nans(name, xla_shape, buf):
+def _check_special(name, xla_shape, buf):
   assert not xla_shape.is_tuple()
   if dtypes.issubdtype(xla_shape.element_type(), np.inexact):
-    if np.any(np.isnan(buf.to_py())):
+    if FLAGS.jax_debug_nans and np.any(np.isnan(buf.to_py())):
       raise FloatingPointError(f"invalid value (nan) encountered in {name}")
+    if FLAGS.jax_debug_infs and np.any(np.isinf(buf.to_py())):
+      raise FloatingPointError(f"invalid value (inf) encountered in {name}")
 
 ### compiling jaxprs
 
@@ -430,11 +437,15 @@ def jaxpr_subcomp(c, jaxpr, backend, axis_env, consts, name_stack, *args):
         source_file=frame.file_name if frame else None,
         source_line=frame.line_num if frame else None))
     in_nodes = _flatmap(read, eqn.invars)
+    # TODO(jakevdp): migrate `translations` table to `translations_with_avals`
     if eqn.primitive in backend_specific_translations[platform]:
       rule = backend_specific_translations[platform][eqn.primitive]
       ans = rule(c, *in_nodes, **eqn.params)
     elif eqn.primitive in translations:
       ans = translations[eqn.primitive](c, *in_nodes, **eqn.params)
+    elif eqn.primitive in translations_with_avals:
+      rule = translations_with_avals[eqn.primitive]
+      ans = rule(c, map(aval, eqn.invars), in_nodes, eqn.params)
     elif eqn.primitive in initial_style_translations:
       new_params = check_backend_params(eqn.params, backend)
       rule = initial_style_translations[eqn.primitive]
@@ -478,10 +489,14 @@ def check_backend_params(params, outer_backend):
   return {k: params[k] for k in params if k != 'backend'}
 
 
-AxisEnv = namedtuple('AxisEnv', ['nreps', 'names', 'sizes', 'devices'])
+class AxisEnv(NamedTuple):
+  """Represents a pmap mesh (only along the replica axes)."""
+  nreps: int
+  names: Tuple[Any, ...]
+  sizes: Tuple[int, ...]
 
-def extend_axis_env(env, name, size):
-  return AxisEnv(env.nreps, env.names + (name,), env.sizes + (size,), env.devices)
+def extend_axis_env(env: AxisEnv, name, size: int):
+  return AxisEnv(env.nreps, env.names + (name,), env.sizes + (size,))
 
 def axis_read(axis_env, axis_name):
   try:
@@ -489,30 +504,38 @@ def axis_read(axis_env, axis_name):
   except ValueError:
     raise NameError("unbound axis name: {}".format(axis_name)) from None
 
-def axis_groups(axis_env, name):
-  if isinstance(name, (list, tuple)):
-    mesh_axes = tuple(unsafe_map(partial(axis_read, axis_env), name))
-  else:
-    mesh_axes = (axis_read(axis_env, name),)
-  return _axis_groups(axis_env.nreps, axis_env.sizes, mesh_axes)
-
-def _axis_groups(nrep, mesh_spec, mesh_axes):
-  trailing_size, ragged = divmod(nrep, prod(mesh_spec))
+def axis_groups(axis_env: AxisEnv, name):
+  if not isinstance(name, (list, tuple)):
+    name = (name,)
+  mesh_axes = tuple(unsafe_map(partial(axis_read, axis_env), name))
+  trailing_size, ragged = divmod(axis_env.nreps, prod(axis_env.sizes))
   assert not ragged
-  full_spec = list(mesh_spec) + [trailing_size]
-  iota = np.arange(prod(full_spec)).reshape(full_spec)
+  mesh_spec = axis_env.sizes + (trailing_size,)
+  return _axis_groups(mesh_spec, mesh_axes)
+
+def _axis_groups(mesh_spec, mesh_axes):
+  """Computes replica group ids for a collective performed over a subset of the mesh.
+
+  Args:
+    mesh_spec: A sequence of integers representing the mesh shape.
+    mesh_axes: A sequence of integers between 0 and `len(mesh_spec)` (exclusive)
+      indicating over which axes the collective is performed.
+  Returns:
+    A tuple of replica groups (i.e. tuples containing replica ids).
+  """
+  iota = np.arange(prod(mesh_spec)).reshape(mesh_spec)
   groups = np.reshape(
       np.moveaxis(iota, mesh_axes, np.arange(len(mesh_axes))),
-      (prod(np.take(full_spec, mesh_axes)), -1))
+      (prod(np.take(mesh_spec, mesh_axes)), -1))
   return tuple(unsafe_map(tuple, groups.T))
 
-def jaxpr_replicas(jaxpr):
+def jaxpr_replicas(jaxpr: core.Jaxpr) -> int:
   """The number of replicas needed for a jaxpr.
 
   For a eqn, multiply the `axis_size` with the `jaxpr_replicas` of the
   subjaxprs. For a list of eqns, take the maximum number of replicas.
   """
-  return max((eqn_replicas(eqn) for eqn in jaxpr.eqns), default=1)
+  return max(unsafe_map(eqn_replicas, jaxpr.eqns), default=1)
 
 # TODO(mattjj): this function assumes that only pmap has a parameter named
 # axis_size, and that it corresponds to cross-replica mapping
@@ -526,10 +549,7 @@ def eqn_replicas(eqn):
     return 1
 
 def initial_style_primitive_replicas(params):
-  nums = (jaxpr_replicas(param if type(param) is core.Jaxpr else param.jaxpr)
-          for param in params.values()
-          if type(param) in (core.Jaxpr, core.ClosedJaxpr))
-  return max(it.chain([1], nums))
+  return max(core.traverse_jaxpr_params(jaxpr_replicas, params), default=1)
 
 # TODO(mattjj,skyewm): the functions here are utilities for checking if
 # not-yet-supported features are used with multi-host programming
@@ -562,7 +582,7 @@ def _xla_call_impl(fun: lu.WrappedFun, *args, device, backend, name, donated_inv
   try:
     return compiled_fun(*args)
   except FloatingPointError:
-    assert FLAGS.jax_debug_nans  # compiled_fun can only raise in this case
+    assert FLAGS.jax_debug_nans or FLAGS.jax_debug_infs  # compiled_fun can only raise in this case
     print("Invalid value encountered in the output of a jit function. "
           "Calling the de-optimized version.")
     # We want to run the wrapped function again (after _xla_callable already ran
@@ -573,7 +593,7 @@ def _xla_call_impl(fun: lu.WrappedFun, *args, device, backend, name, donated_inv
     # by any transformation_with_aux's applied to fun. Since this is
     # intentional here, to avoid "Store occupied" errors we reset the stores to
     # be empty.
-    for store in fun.stores: store.reset()
+    for store in fun.stores: store and store.reset()
     return fun.call_wrapped(*args)  # probably won't return
 
 def flatten_shape(s: XlaShape) -> Sequence[Tuple[Sequence[int], XlaShape]]:
@@ -686,7 +706,7 @@ def _xla_callable(fun: lu.WrappedFun, device, backend, name, donated_invars, *ar
   xla_consts = _xla_consts(c, consts)
   xla_args, donated_invars = _xla_callable_args(c, abstract_args, tuple_args, donated_invars=donated_invars)
   out_nodes = jaxpr_subcomp(
-      c, jaxpr, backend, AxisEnv(nreps, (), (), None), xla_consts,
+      c, jaxpr, backend, AxisEnv(nreps, (), ()), xla_consts,
       extend_name_stack(wrap_name(name, 'jit')), *xla_args)
   out_tuple = xops.Tuple(c, out_nodes)
   backend = xb.get_backend(backend)
@@ -759,20 +779,24 @@ def _xla_callable_device(nreps, backend, device, arg_devices):
 _replicated_param = object()
 
 def _xla_callable_args(
-    c, avals, tuple_args, replicated=None,
-    partitions: Optional[Sequence[Optional[Sequence[int]]]] = None,
-    donated_invars = None):
+    c, avals, tuple_args, *,
+    replicated=None,
+    partitions=None,
+    partitions_proto: bool = False,
+    donated_invars=None):
   assert partitions is None or len(partitions) == len(avals)
   if not tuple_args:
     if replicated is None:
       replicated = [None] * len(avals)
     if partitions is None:
       parts: List[object] = [None] * len(avals)
+    elif partitions_proto:
+      parts = partitions
     else:
       parts = [_replicated_param if part is None else part
                for part in partitions]
     counts = it.count()
-    xla_args = [_xla_param(c, next(counts), xla_shape, r, p)
+    xla_args = [_xla_param(c, next(counts), xla_shape, r, p, partitions_proto)
                 if a is not abstract_token else xops.CreateToken(c)
                 for (a, r, p) in safe_zip(avals, replicated, parts)
                 for xla_shape in aval_to_xla_shapes(a)]
@@ -785,31 +809,37 @@ def _xla_callable_args(
     if replicated is not None:
       replicated = [r for a, r in zip(avals, replicated)
                     if a is not abstract_token]
-    tuple_parts = tuple(partitions) if partitions is not None else None
+    if partitions is None:
+      tuple_parts = None
+    elif partitions_proto:
+      tuple_parts = xb.tuple_sharding_proto(partitions)
+    else:
+      tuple_parts = tuple(partitions)
     tuple_shape = xc.Shape.tuple_shape(
         [shape for a in avals for shape in aval_to_xla_shapes(a) if a is not abstract_token])
-    tuple_param = _xla_param(c, 0, tuple_shape, replicated, tuple_parts)
+    tuple_param = _xla_param(c, 0, tuple_shape, replicated, tuple_parts, partitions_proto)
     xla_inputs = iter(xla_destructure(c, tuple_param))
     xla_args = [next(xla_inputs) if a is not abstract_token else
                 xops.CreateToken(c) for a in avals]
     assert next(xla_inputs, None) is None
     return xla_args, donated_invars
 
-def _xla_param(builder, param_num, xla_shape, replicated, partitions):
+def _xla_param(builder, param_num, xla_shape, replicated, partitions, parts_proto):
   make_param = partial(xb.parameter, builder, param_num, xla_shape,
                        replicated=replicated)
+  with_sharding = xb.with_sharding_proto if parts_proto else xb.with_sharding
   if partitions is None:
     return make_param()
   elif partitions is _replicated_param:
-    return xb.with_sharding(builder, None, make_param)
+    return with_sharding(builder, None, make_param)
   else:
-    return xb.with_sharding(builder, partitions, make_param)
+    return with_sharding(builder, partitions, make_param)
 
 def _execute_compiled(compiled: XlaExecutable, avals, handlers, *args):
   device, = compiled.local_devices()
   input_bufs = list(it.chain.from_iterable(device_put(x, device) for x in args if x is not token))
   out_bufs = compiled.execute(input_bufs)
-  if FLAGS.jax_debug_nans: check_nans(xla_call_p, out_bufs)
+  check_special(xla_call_p, out_bufs)
   return [handler(*bs) for handler, bs in zip(handlers, _partition_outputs(avals, out_bufs))]
 
 def _execute_replicated(compiled: XlaExecutable, avals, handlers, *args):
@@ -817,7 +847,7 @@ def _execute_replicated(compiled: XlaExecutable, avals, handlers, *args):
       list(it.chain.from_iterable(device_put(x, device) for x in args if x is not token))
       for device in compiled.local_devices()]
   out_bufs = compiled.execute_on_local_devices(input_bufs)[0]
-  if FLAGS.jax_debug_nans: check_nans(xla_call_p, out_bufs)
+  check_special(xla_call_p, out_bufs)
   return [handler(*bs) for handler, bs in zip(handlers, _partition_outputs(avals, out_bufs))]
 
 def _execute_trivial(jaxpr, device: Optional[Device], consts, avals, handlers, *args):
@@ -878,6 +908,7 @@ ad.primitive_transposes[xla_call_p] = partial(ad.call_transpose, xla_call_p)
 ### translation tables
 
 translations: Dict[core.Primitive, Callable] = {}
+translations_with_avals: Dict[core.Primitive, Callable] = {}
 parallel_translations: Dict[core.Primitive, Callable] = {}
 initial_style_translations: Dict[core.Primitive, Callable] = {}
 call_translations: Dict[core.Primitive, Callable] = {}
@@ -906,22 +937,18 @@ def _tuple_output(*args, **kwargs):
   ans = yield args, kwargs
   yield (ans,)
 
-def lower_fun(fun, multiple_results, parallel=False):
-  # This function can only be used to lower functions that take JAX array types
-  # as arguments (and e.g. don't accept unit values), because it assumes it can
-  # map from XLA types to JAX types. In general that mapping is not possible (as
-  # the mapping from JAX types to XLA types is not invertible), but for now at
-  # least we assume that the mapping from JAX *array* types to XLA array types
-  # is invertible. This assumption is unchecked!
-  # TODO(mattjj): remove assumption can map XLA array types to JAX array types
+def lower_fun(fun, multiple_results, parallel=False, with_avals=False):
+  # TODO(jakevdp): migrate dependent code & always use the with_avals=True.
   def f(c, *xla_args, **params):
-    # TODO(mattjj): revise this 'calling convention'
     avals = [_array_aval_from_xla_shape(c.get_shape(x)) for x in xla_args]
+    return f_with_avals(c, avals, xla_args, params)
+
+  def f_with_avals(c, avals, xla_args, params):
     if parallel:
       axis_env = params.pop('axis_env')
       del params['platform']
     else:
-      axis_env = AxisEnv(1, (), (), None)
+      axis_env = AxisEnv(1, (), ())
     wrapped_fun = lu.wrap_init(fun, params)
     if not multiple_results:
       wrapped_fun = _tuple_output(wrapped_fun)
@@ -935,12 +962,13 @@ def lower_fun(fun, multiple_results, parallel=False):
                                           stage_out=True)  # type: ignore
       xla_consts = _xla_consts(c, consts)
       outs = jaxpr_subcomp(c, jaxpr, None, axis_env, xla_consts, '', *xla_args)
-    if multiple_results:
+    if multiple_results or any(v.aval._num_buffers > 1 for v in jaxpr.outvars):
       return xops.Tuple(c, outs)
     else:
       assert len(outs) == 1, outs
       return outs[0]
-  return f
+
+  return f_with_avals if with_avals else f
 
 def _array_aval_from_xla_shape(xla_shape):
   # This function instantiates the assumption that we can map fro XLA array
@@ -983,15 +1011,39 @@ def _forward_method(attrname, self, fun, *args):
 _forward_to_value = partial(_forward_method, "_value")
 
 
-def make_device_array(aval: core.ShapedArray, device: Optional[Device],
-                      lazy_expr: lazy.LazyExpr, device_buffer: PyLocalBuffer):
+# The following is used for the type _CppDeviceArray or _DeviceArray.
+DeviceArrayProtocol = Any
+if hasattr(xc, "DeviceArrayBase"):
+  DeviceArray = xc.DeviceArrayBase
+else:
+  # prior to jaxlib version 0.1.58.
+  class DeviceArray:  # type: ignore
+    pass
+
+_CppDeviceArray: DeviceArrayProtocol = xc.Buffer
+
+_EXPERIMENTAL_CPP_DEVICE_ARRAY = False
+
+
+def make_device_array(
+    aval: core.ShapedArray,
+    device: Optional[Device],
+    lazy_expr: Optional[lazy.LazyExpr],
+    device_buffer: Union[PyLocalBuffer, "DeviceConstant"],
+) -> Union[PyLocalBuffer, "_DeviceArray"]:
   """Returns a DeviceArray implementation based on arguments.
 
-  External users should not call the `DeviceArray` constructor as it's an
-  internal implementation detail. This is added to smooth the transition to a
-  C++ equivalent implementation.
+  This is to be used only within JAX. It will return either a PythonDeviceArray
+  or a C++ equivalent implementation.
   """
-  return DeviceArray(aval, device, lazy_expr, device_buffer)
+  if (_EXPERIMENTAL_CPP_DEVICE_ARRAY and lazy.is_trivial(lazy_expr) and
+      not isinstance(device_buffer, DeviceConstant)):
+    assert isinstance(device_buffer, _CppDeviceArray)
+    device_buffer._device = device    # pylint: disable=protected-access
+    device_buffer.aval = aval
+    return device_buffer
+
+  return _DeviceArray(aval, device, lazy_expr, device_buffer)
 
 
 def type_is_device_array(x):
@@ -999,17 +1051,16 @@ def type_is_device_array(x):
 
   Use this function instead of `type(x) is Devicearray`.
   """
-  # TODO(jblespiau): Extend to also support `PyBuffer`.
-  return type(x) is DeviceArray
+  type_x = type(x)
+  return type_x is _DeviceArray or type_x is _CppDeviceArray
 
 
-class DeviceArray:
+class _DeviceArray(DeviceArray):  # type: ignore
   """A DeviceArray is an ndarray backed by a single device memory buffer."""
   # We don't subclass ndarray because that would open up a host of issues,
   # but lax_numpy.py overrides isinstance behavior and attaches ndarray methods.
   __slots__ = [
-      "aval", "device_buffer", "_npy_value", "_device", "_lazy_expr",
-      "__weakref__"
+      "aval", "device_buffer", "_npy_value", "_device", "_lazy_expr"
   ]
   __array_priority__ = 100
 
@@ -1030,6 +1081,7 @@ class DeviceArray:
         `LazyExpr`.
       device_buffer: The underlying buffer owning the on-device data.
     """
+    DeviceArray.__init__(self)
     self.aval = aval
     self.device_buffer = device_buffer
     self._device = device
@@ -1086,10 +1138,6 @@ class DeviceArray:
   def ndim(self):
     return len(self.aval.shape)
 
-  def copy(self):
-    """Returns an ndarray (backed by host memory, not device memory)."""
-    return np.asarray(self)
-
   def copy_to_host_async(self):
     """Requests a copy of the buffer to the host."""
     self._check_if_deleted()
@@ -1111,9 +1159,24 @@ class DeviceArray:
     self.device_buffer = deleted_buffer
     self._npy_value = None
 
+  @property
+  def __cuda_array_interface__(self):
+    return _force(self).device_buffer.__cuda_array_interface__
+
+
+# Adding methods dynamically to both _DeviceArray and _CppDeviceArray
+# pylint: disable=protected-access
+for device_array in [_DeviceArray, _CppDeviceArray]:
+
+
+  def copy(self):
+    """Returns an ndarray (backed by host memory, not device memory)."""
+    return np.asarray(self)
+  setattr(device_array, "copy", copy)
+
   def __repr__(self):
-    line_width = np.get_printoptions()['linewidth']
-    prefix = '{}('.format(self.__class__.__name__)
+    line_width = np.get_printoptions()["linewidth"]
+    prefix = '{}('.format(self.__class__.__name__.lstrip('_'))
     s = np.array2string(self._value, prefix=prefix, suffix=',',
                         separator=', ', max_line_width=line_width)
     dtype_str = 'dtype={})'.format(self.dtype.name)
@@ -1123,6 +1186,8 @@ class DeviceArray:
       sep = ' ' * len(prefix)
     return "{}{},{}{}".format(prefix, s, sep, dtype_str)
 
+  setattr(device_array, "__repr__", __repr__)
+
   def item(self):
     if dtypes.issubdtype(self.dtype, np.complexfloating):
       return complex(self)
@@ -1135,11 +1200,15 @@ class DeviceArray:
     else:
       raise TypeError(self.dtype)
 
+  setattr(device_array, "item", item)
+
   def __len__(self):
     try:
       return self.aval.shape[0]
     except IndexError as err:
-      raise TypeError("len() of unsized object") from err # same as numpy error
+      raise TypeError("len() of unsized object") from err  # same as numpy error
+
+  setattr(device_array, "__len__", __len__)
 
   def __iter__(self):
     if self.ndim == 0:
@@ -1147,12 +1216,16 @@ class DeviceArray:
     else:
       return self._value.__iter__()
 
+  setattr(device_array, "__iter__", __iter__)
+
   def __reversed__(self):
     if self.ndim == 0:
       raise TypeError("iteration over a 0-d array")
     else:
       return reversed(self._value)
 
+  setattr(device_array, "__reversed__", __reversed__)
+
   def __format__(self, format_spec):
     # Simulates behavior of https://github.com/numpy/numpy/pull/9883
     if self.ndim == 0:
@@ -1160,38 +1233,46 @@ class DeviceArray:
     else:
       return format(self._value, format_spec)
 
+  setattr(device_array, "__format__", __format__)
+
   def __array__(self, dtype=None, context=None):
     return np.asarray(self._value, dtype=dtype)
 
-  @property
-  def __cuda_array_interface__(self):
-    return _force(self).device_buffer.__cuda_array_interface__
+  setattr(device_array, "__array__", __array__)
 
-  __str__ = partialmethod(_forward_to_value, str)
-  __bool__ = __nonzero__ = partialmethod(_forward_to_value, bool)
-  def __float__(self): return self._value.__float__()
-  def __int__(self): return self._value.__int__()
-  def __complex__(self): return self._value.__complex__()
-  __hex__ = partialmethod(_forward_to_value, hex)
-  __oct__ = partialmethod(_forward_to_value, oct)
-  __index__ = partialmethod(_forward_to_value, op.index)
-  def tobytes(self, order='C'): return self._value.tobytes(order)
-  def tolist(self): return self._value.tolist()
+  setattr(device_array, "__str__", partialmethod(_forward_to_value, str))
+  setattr(device_array, "__bool__", partialmethod(_forward_to_value, bool))
+  setattr(device_array, "__nonzero__", partialmethod(_forward_to_value, bool))
+  setattr(device_array, "__float__", lambda self: self._value.__float__())
+  setattr(device_array, "__int__", lambda self: self._value.__int__())
+  setattr(device_array, "__complex__", lambda self: self._value.__complex__())
+  setattr(device_array, "__hex__", partialmethod(_forward_to_value, hex))
+  setattr(device_array, "__oct__", partialmethod(_forward_to_value, oct))
+  setattr(device_array, "__index__", partialmethod(_forward_to_value, op.index))
+  to_bytes = lambda self, order="C": self._value.tobytes(order)
+  setattr(device_array, "tobytes", to_bytes)
+  del to_bytes
+  setattr(device_array, "tolist", lambda self: self._value.tolist())
 
   # pickle saves and loads just like an ndarray
-  __reduce__ = partialmethod(_forward_to_value, op.methodcaller("__reduce__"))
+  setattr(device_array, "__reduce__",
+          partialmethod(_forward_to_value, op.methodcaller("__reduce__")))
 
   # clobbered when jax.numpy is imported, but useful in tests
-  def __eq__(self, other): return self._value == other
+  setattr(device_array, "__eq__", lambda self, other: self._value == other)
 
   def __hash__(self):
     raise TypeError("JAX DeviceArray, like numpy.ndarray, is not hashable.")
 
+  setattr(device_array, "__hash__", __hash__)
+
   # The following methods are dynamically overridden in lax_numpy.py.
-  def __getitem__(self, i): raise NotImplementedError
+  def raise_not_implemented():
+    raise NotImplementedError
+
+  setattr(device_array, "__getitem__", lambda self, i: raise_not_implemented())
+# pylint: enable=protected-access
 
-_DeviceArray = DeviceArray
-_CppDeviceArray = xc.Buffer
 
 class DeletedBuffer(object): pass
 deleted_buffer = DeletedBuffer()
@@ -1205,10 +1286,11 @@ class DeviceConstant(object):
 def is_device_constant(x):
   return type_is_device_array(x) and type(x.device_buffer) is DeviceConstant
 
-core.literalable_types.add(DeviceArray)
-core.pytype_aval_mappings[DeviceArray] = ConcreteArray
-pytype_aval_mappings[DeviceArray] = op.attrgetter('aval')
-canonicalize_dtype_handlers[DeviceArray] = identity
+for device_array in [_CppDeviceArray, _DeviceArray]:
+  core.literalable_types.add(device_array)
+  core.pytype_aval_mappings[device_array] = ConcreteArray
+  pytype_aval_mappings[device_array] = op.attrgetter('aval')
+  canonicalize_dtype_handlers[device_array] = identity
 
 def _device_array_constant_handler(c, val, canonicalize_types=True):
   if is_device_constant(val):
@@ -1216,20 +1298,23 @@ def _device_array_constant_handler(c, val, canonicalize_types=True):
   else:
     base_val = xb.constant(c, val.device_buffer.to_py())
     return lazy.stage_lexpr(c, val._lazy_expr, base_val)
-xb.register_constant_handler(DeviceArray, _device_array_constant_handler)
+xb.register_constant_handler(_DeviceArray, _device_array_constant_handler)
+xb.register_constant_handler(_CppDeviceArray, _device_array_constant_handler)
 
-def _device_put_device_array(x: DeviceArray, device: Optional[Device]):
+def _device_put_device_array(x: Union[DeviceArrayProtocol, _DeviceArray], device: Optional[Device]):
   x = _copy_device_array_to_device(x, device)
   return (_force(x).device_buffer,)
-device_put_handlers[DeviceArray] = _device_put_device_array
+device_put_handlers[_CppDeviceArray] = _device_put_device_array
+device_put_handlers[_DeviceArray] = _device_put_device_array
 
-def _copy_device_array_to_device(x: DeviceArray, device: Optional[xc.Device]) -> DeviceArray:
+def _copy_device_array_to_device(x: Union[DeviceArrayProtocol, _DeviceArray], device: Optional[xc.Device]) -> Union[DeviceArrayProtocol, _DeviceArray]:
   if device is None:
     # no copying to be done because there's no target specified
     return x
   elif is_device_constant(x):
     # create a new DeviceArray with the same lazy expr, no copying
-    return DeviceArray(x.aval, device, x._lazy_expr, DeviceConstant(device))
+    return make_device_array(x.aval, device, x._lazy_expr,
+                               DeviceConstant(device))
   elif xb.get_device_backend(device).platform == x.device_buffer.platform():
     # source and target platforms are the same
     if x.device_buffer.device() == device:
@@ -1245,9 +1330,9 @@ def _copy_device_array_to_device(x: DeviceArray, device: Optional[xc.Device]) ->
     # buffers from different XLA backends are passed through the host.
     backend = xb.get_device_backend(device)
     moved_buf = backend.buffer_from_pyval(x.device_buffer.to_py(), device)
-  return DeviceArray(x.aval, device, x._lazy_expr, moved_buf)
+  return _DeviceArray(x.aval, device, x._lazy_expr, moved_buf)
 
-def _force(x: DeviceArray) -> DeviceArray:
+def _force(x: DeviceArrayProtocol) -> DeviceArrayProtocol:
   if lazy.is_trivial(x._lazy_expr):
     return x
   else:
@@ -1258,12 +1343,12 @@ def _force(x: DeviceArray) -> DeviceArray:
       device = x.device_buffer.device()
     force_fun = _lazy_force_computation(x.aval, device, x._lazy_expr)
     result = force_fun(x)
-    return DeviceArray(x.aval, x._device, lazy.array(x.aval.shape), result)
+    return make_device_array(x.aval, x._device, lazy.array(x.aval.shape), result)
 
 @cache()
 def _lazy_force_computation(aval: core.ShapedArray,
                             device: Device, lexpr: lazy.LazyExpr
-                            ) -> Callable[[DeviceArray], PyLocalBuffer]:
+                            ) -> Callable[[_DeviceArray], PyLocalBuffer]:
   c = xb.make_computation_builder("lazy_force")
   if lazy.is_constant(lexpr):
     param = None
@@ -1283,7 +1368,7 @@ def _lazy_force_computation(aval: core.ShapedArray,
       device_assignment=device and (device.id,))
   compiled = backend_compile(xb.get_device_backend(device), built_c, options)
 
-  force_fun: Callable[[DeviceArray], DeviceArray]
+  force_fun: Callable[[_DeviceArray], PyLocalBuffer]
   if lazy.is_constant(lexpr):
     def force_fun(_):
       return compiled.execute([])[0]
@@ -1308,7 +1393,7 @@ device_put_p = core.Primitive('device_put')
 device_put_p.def_impl(_device_put_impl)
 device_put_p.def_abstract_eval(lambda x, device=None: x)
 translations[device_put_p] = lambda c, x, device=None: x
-ad.deflinear(device_put_p, lambda cotangent, **kwargs: [cotangent])
+ad.deflinear2(device_put_p, lambda cotangent, _, **kwargs: [cotangent])
 masking.defvectorized(device_put_p)
 
 
@@ -1353,7 +1438,7 @@ def _remat_translation_rule(c, axis_env, in_nodes,
   dummy_subc = dummy_subc.build(xops.Tuple(dummy_subc, out_nodes))
 
   return xops.Conditional(pred, true_op, remat_subc, false_op, dummy_subc)
-call_translations[pe.remat_call_p] = _remat_translation_rule
+call_translations[pe.remat_call_p] = _remat_translation_rule  # type: ignore
 
 
 ad.primitive_transposes[core.named_call_p] = partial(ad.call_transpose,
