@@ -63,12 +63,18 @@ data must be immutable, because it will be stored in function memoization tables
 """
 
 import threading
+from functools import partial
 from typing import Any, Tuple, Callable
 import weakref
 
-from .util import curry
+from . import core
+from ._src.util import curry
+from .tree_util import tree_map
 
 from ._src import traceback_util
+
+from .config import FLAGS
+
 traceback_util.register_exclusion(__file__)
 
 
@@ -109,7 +115,7 @@ class Store(object):
 class WrappedFun(object):
   """Represents a function `f` to which `transforms` are to be applied.
 
-  Arguments:
+  Args:
     f: the function to be transformed.
     transforms: a list of `(gen, gen_static_args)` tuples representing
       transformations to apply to `f.` Here `gen` is a generator function
@@ -154,7 +160,7 @@ class WrappedFun(object):
       gen = gen(*(gen_static_args + tuple(args)), **kwargs)
       args, kwargs = next(gen)
       stack.append((gen, out_store))
-    gen = None
+    gen = gen_static_args = out_store = None
 
     try:
       ans = self.f(*args, **dict(self.params, **kwargs))
@@ -167,7 +173,7 @@ class WrappedFun(object):
         stack.pop()[0].close()
       raise
 
-    del args
+    args = kwargs = None
     while stack:
       gen, out_store = stack.pop()
       ans = gen.send(ans)
@@ -242,7 +248,10 @@ def cache(call: Callable):
 
   def memoized_fun(fun: WrappedFun, *args):
     cache = fun_caches.setdefault(fun.f, {})
-    key = (fun.transforms, fun.params, args)
+    if core.debug_state.check_leaks:
+      key = (_copy_main_traces(fun.transforms), fun.params, args, bool(FLAGS.jax_enable_x64))
+    else:
+      key = (fun.transforms, fun.params, args, bool(FLAGS.jax_enable_x64))
     result = cache.get(key, None)
     if result is not None:
       ans, stores = result
@@ -266,6 +275,13 @@ def cache(call: Callable):
 
   return memoized_fun
 
+@partial(partial, tree_map)
+def _copy_main_traces(x):
+  if isinstance(x, core.MainTrace):
+    return core.MainTrace(x.level, x.trace_type, **x.payload)
+  else:
+    return x
+
 
 @transformation
 def hashable_partial(x, *args):
